from typing import TYPE_CHECKING, Any

from prefect.storage import Storage
from prefect.utilities.storage import extract_flow_from_file

if TYPE_CHECKING:
    from prefect.core.flow import Flow


class CodeCommit(Storage):
    """
    CodeCommit storage class. This class represents the Storage interface for Flows stored
    in `.py` files in a CodeCommit repository.

    This class represents a mapping of flow name to file paths contained in the git repo,
    meaning that all flow files should be pushed independently. A typical workflow using
    this storage type might look like the following:

    - Compose flow `.py` file where flow has CodeCommit storage:

    ```python
    flow = Flow("my-flow")
    flow.storage = CodeCommit(repo="my/repo", path="/flows/flow.py")
    ```

    - Push this `flow.py` file to the `my/repo` repository under `/flows/flow.py`.

    - Call `prefect register -f flow.py` to register this flow with CodeCommit storage.

    Args:
        - repo (str): the name of a CodeCommit repository to store this Flow
        - path (str, optional): a path pointing to a flow file in the repo
        - commit (str, optional): fully quaified reference that identifies the commit
          that contains the file. For example, you can specify a full commit ID, a tag,
          a branch name, or a reference such as refs/heads/master. If none is provided,
          the head commit is used
        - client_options (dict, optional): Additional options for the `boto3` client.
        - **kwargs (Any, optional): any additional `Storage` initialization options
    """

    def __init__(
        self,
        repo: str,
        path: str = None,
        commit: str = None,
        client_options: dict = None,
        **kwargs: Any
    ) -> None:
        self.repo = repo
        self.path = path
        self.commit = commit
        self.client_options = client_options

        super().__init__(**kwargs)

    def get_flow(self, flow_name: str) -> "Flow":
        """
        Given a flow name within this Storage object, load and return the Flow.

        Args:
            - flow_name (str): the name of the flow to return.

        Returns:
            - Flow: the requested flow
        """
        if flow_name not in self.flows:
            raise ValueError("Flow is not contained in this Storage")
        flow_location = self.flows[flow_name]

        client = self._boto3_client

        try:
            file_contents = client.get_file(
                repositoryName=self.repo,
                commitSpecifier=self.commit,
                filePath=flow_location,
            )["fileContent"]
            decoded_contents = file_contents.decode("utf-8")
        except Exception as exc:
            self.logger.error(
                "Error retrieving file contents from {} on repo {}. Ensure the file exists.".format(
                    flow_location, self.repo
                )
            )
            raise exc

        return extract_flow_from_file(
            file_contents=decoded_contents, flow_name=flow_name
        )

    def add_flow(self, flow: "Flow") -> str:
        """
        Method for storing a new flow as bytes in the local filesytem.

        Args:
            - flow (Flow): a Prefect Flow to add

        Returns:
            - str: the location of the added flow in the repo

        Raises:
            - ValueError: if a flow with the same name is already contained in this storage
        """
        if flow.name in self:
            raise ValueError(
                'Name conflict: Flow with the name "{}" is already present in this storage.'.format(
                    flow.name
                )
            )

        self.flows[flow.name] = self.path  # type: ignore
        self._flows[flow.name] = flow
        return self.path  # type: ignore

    @property
    def _boto3_client(self):  # type: ignore
        from prefect.utilities.aws import get_boto_client

        kwargs = self.client_options or {}
        return get_boto_client(resource="codecommit", credentials=None, **kwargs)



import json

try:
    import importlib.resources as pkg_resources
except ImportError:
    import importlib_resources as pkg_resources

import boto3
from botocore.exceptions import WaiterError
from botocore.waiter import WaiterModel, create_waiter_with_client

from prefect.tasks.aws import waiters
from prefect import Task
from prefect.engine.signals import FAIL
from prefect.utilities.aws import get_boto_client
from prefect.utilities.tasks import defaults_from_attrs


class AWSClientWait(Task):
    """
    Task for waiting on a long-running AWS job. Uses the underlying boto3 waiter functionality.

    For authentication, there are two options: you can set the `AWS_CREDENTIALS` Prefect Secret
    containing your AWS access keys which will be passed directly to the `boto3` client, or you
    can [configure your flow's runtime
    environment](https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html#guide-configuration)
    for `boto3`.

    Args:
        - client (str, optional): The AWS client on which to wait (e.g., 'batch', 'ec2', etc)
        - waiter_name (str, optional): The name of the waiter to instantiate. Can be a boto-supported
            waiter or one of prefect's custom waiters. Currently, prefect offers three additional
            waiters for AWS Batch: `"JobExists"` waits for a job to be instantiated, `"JobRunning"`
            waits for a job to start running, and `"JobComplete"` waits for a job to finish. You can
            find the definitions for all prefect-defined waiters [here](https://github.com/PrefectHQ/prefect/tree/master/src/prefect/tasks/aws/waiters).  # noqa
            You may also use a custom waiter name, if you supply an accompanying waiter definition
            dict.
        - waiter_definition (dict, optional): A valid custom waiter model, as a dict. Note that if
            you supply a custom definition, it is assumed that the provided 'waiter_name' is
            contained within the waiter definition dict.
        - boto_kwargs (dict, optional): additional kekyword arguments to forward to the boto client.
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        client: str = None,
        waiter_name: str = None,
        waiter_definition: dict = None,
        boto_kwargs: dict = None,
        **kwargs,
    ):
        self.client = client
        self.waiter_name = waiter_name
        self.waiter_definition = waiter_definition
        self.boto_kwargs = {} if boto_kwargs is None else boto_kwargs

        super().__init__(**kwargs)

    @defaults_from_attrs("client", "waiter_name", "waiter_definition")
    def run(
        self,
        client: str = None,
        waiter_name: str = None,
        waiter_definition: dict = None,
        waiter_kwargs: dict = None,
        credentials: str = None,
    ):
        """
        Task for waiting on a long-running AWS job. Uses the underlying boto3 waiter functionality.

        Args:
            - client (str): The AWS client on which to wait (e.g., 'batch', 'ec2', etc)
            - waiter_name (str, optional): The name of the waiter to instantiate. Can be a boto-supported
                waiter or one of prefect's custom waiters. Currently, prefect offers three additional
                waiters for AWS Batch: `"JobExists"` waits for a job to be instantiated, `"JobRunning"`
                waits for a job to start running, and `"JobComplete"` waits for a job to finish. You can
                find the definitions for all prefect-defined waiters [here](https://github.com/PrefectHQ/prefect/tree/master/src/prefect/tasks/aws/waiters).  # noqa
                You may also use a custom waiter name, if you supply an accompanying waiter definition
                dict.
            - waiter_definition (dict, optional): A valid custom waiter model, as a dict. Note that if
                you supply a custom definition, it is assumed that the provided 'waiter_name' is
                contained within the waiter definition dict.
            - waiter_kwargs (dict, optional): Arguments to pass to the `waiter.wait(...)` method. Will
                depend upon the specific waiter being called.
            - credentials (dict, optional): your AWS credentials passed from an upstream
                Secret task; this Secret must be a JSON string
                with two keys: `ACCESS_KEY` and `SECRET_ACCESS_KEY` which will be
                passed directly to `boto3`.  If not provided here or in context, `boto3`
                will fall back on standard AWS rules for authentication.
        """
        if not client:
            raise ValueError("An AWS client string must be provided.")

        if not waiter_name:
            raise ValueError("A waiter name must be provided.")

        if not waiter_kwargs:
            waiter_kwargs = {}

        boto_client = get_boto_client(
            client, credentials=credentials, **self.boto_kwargs
        )
        if waiter_definition:
            # Use user-provided waiter definition
            waiter_model = WaiterModel(waiter_definition)
            waiter = create_waiter_with_client(waiter_name, waiter_model, boto_client)
        else:
            # Use either boto-provided or prefect-provided waiter
            if waiter_name in boto_client.waiter_names:
                waiter = boto_client.get_waiter(waiter_name)
            else:
                waiter = self._load_prefect_waiter(boto_client, client, waiter_name)

        try:
            waiter.wait(**waiter_kwargs)
        except WaiterError as e:
            raise FAIL(
                f"AWS {client} waiter '{waiter_name}' failed with: {str(e)}"
            ) from e

    @staticmethod
    def _load_prefect_waiter(
        boto_client: "boto3.client", client_str: str, waiter_name: str
    ):
        """
        Load a custom waiter from the ./waiters directory.
        """
        try:
            # Instantiate waiter from accompanying client json file
            with pkg_resources.open_text(waiters, f"{client_str}.json") as handle:
                waiter_model = WaiterModel(json.load(handle))

            return create_waiter_with_client(waiter_name, waiter_model, boto_client)
        except Exception as err:
            raise ValueError(
                f"Unable to load waiter '{waiter_name}' for AWS client '{client_str}'."
            ) from err



from typing import Any, TYPE_CHECKING

from prefect.storage.base import Storage
from prefect.utilities.storage import extract_flow_from_module

if TYPE_CHECKING:
    from prefect import Flow


class Module(Storage):
    """
    A Prefect Storage class for referencing flows that can be imported from a
    python module.

    Args:
        - module (str): The module to import the flow from.
        - **kwargs (Any, optional): any additional `Storage` options.

    Example:

    Suppose you have a python module `myproject.flows` that contains all your
    Prefect flows. If this module is installed and available in your execution
    environment you can use `Module` storage to reference and load the flows.

    ```python
    from prefect import Flow
    from prefect.storage import Module

    flow = Flow("module storage example")
    flow.storage = Module("myproject.flows")

    # Tip: you can use `__name__` to automatically reference the current module
    flow.storage = Module(__name__)
    ```
    """

    def __init__(self, module: str, **kwargs: Any) -> None:
        self.module = module
        super().__init__(**kwargs)

    def get_flow(self, flow_name: str) -> "Flow":
        """
        Given a flow name within this Storage object, load and return the Flow.

        Args:
            - flow_name (str): the name of the flow to return.

        Returns:
            - Flow: the requested flow
        """
        if flow_name not in self.flows:
            raise ValueError("Flow is not contained in this Storage")
        return extract_flow_from_module(module_str=self.module, flow_name=flow_name)

    def add_flow(self, flow: "Flow") -> str:
        """
        Add a new flow to this storage object.

        Args:
            - flow (Flow): the Prefect Flow to add.

        Returns:
            - str: the location of the newly added flow in this Storage object
        """
        if flow.name in self:
            raise ValueError(
                'Name conflict: Flow with the name "{}" is already present in this storage.'.format(
                    flow.name
                )
            )

        self.flows[flow.name] = self.module
        self._flows[flow.name] = flow
        return self.module



import os
from copy import deepcopy
from typing import Iterable, Dict, Any

import slugify
import yaml

from prefect import config
from prefect.agent import Agent
from prefect.run_configs import ECSRun
from prefect.utilities.agent import get_flow_image, get_flow_run_command
from prefect.utilities.filesystems import read_bytes_from_path
from prefect.utilities.graphql import GraphQLResult


DEFAULT_TASK_DEFINITION_PATH = os.path.join(
    os.path.dirname(__file__), "task_definition.yaml"
)


def merge_run_task_kwargs(opts1: dict, opts2: dict) -> dict:
    """Merge two `run_task_kwargs` dicts, given precedence to `opts2`.

    Values are merged with the following heuristics:

    - Anything outside of `overrides.containerOverrides` is merged directly,
      with precedence given to `opts2`
    - Dicts in the `overrides.containerOverrides` list are matched on their
      `"name"` fields, then merged directly (with precedence given to `opts2`).

    Args:
        - opts1 (dict): A dict of kwargs for `run_task`
        - opts2 (dict): A second dict of kwargs for `run_task`.

    Returns:
        - dict: A merged dict of kwargs
    """

    out = deepcopy(opts1)

    # Everything except 'overrides' merge directly
    for k, v in opts2.items():
        if k != "overrides":
            out[k] = v

    # Everything in `overrides` except `containerOverrides` merge directly
    overrides = opts2.get("overrides", {})
    if overrides:
        out_overrides = out.setdefault("overrides", {})
        for k, v in overrides.items():
            if k != "containerOverrides":
                out_overrides[k] = v

    # Entries in `containerOverrides` are paired by name, and then merged
    container_overrides = overrides.get("containerOverrides")
    if container_overrides:
        out_container_overrides = out_overrides.setdefault("containerOverrides", [])
        for entry in container_overrides:
            for out_entry in out_container_overrides:
                if out_entry.get("name") == entry.get("name"):
                    out_entry.update(entry)
                    break
            else:
                out_container_overrides.append(entry)
    return out


class ECSAgent(Agent):
    """
    Agent which deploys flow runs as ECS tasks.

    Args:
        - agent_config_id (str, optional): An optional agent configuration ID
            that can be used to set configuration based on an agent from a
            backend API. If set all configuration values will be pulled from
            the backend agent configuration.
        - name (str, optional): An optional name to give this agent. Can also
            be set through the environment variable `PREFECT__CLOUD__AGENT__NAME`.
            Defaults to "agent".
        - labels (List[str], optional): A list of labels, which are arbitrary
            string identifiers used by Prefect Agents when polling for work.
        - env_vars (dict, optional): A dictionary of environment variables and
            values that will be set on each flow run that this agent submits
            for execution.
        - max_polls (int, optional): Maximum number of times the agent will
            poll Prefect Cloud for flow runs; defaults to infinite.
        - agent_address (str, optional):  Address to serve internal api at.
            Currently this is just health checks for use by an orchestration
            layer. Leave blank for no api server (default).
        - no_cloud_logs (bool, optional): Disable logging to a Prefect backend
            for this agent and all deployed flow runs. Defaults to `False`.
        - task_definition_path (str, optional): Path to a task definition
            template to use when defining new tasks. If not provided, the
            default template will be used.
        - run_task_kwargs_path (str, optional): Path to a `yaml` file
            containing default kwargs to pass to `ECS.client.run_task`. May be
            a local path, or a remote path on e.g. `s3`.
        - aws_access_key_id (str, optional): AWS access key id for connecting
            the boto3 client. If not provided, will be loaded from your
            environment (via either the `AWS_ACCESS_KEY_ID` environment
            variable, or the `~/.aws/config` file). See
            [the boto3 credentials docs][1] for more information.
        - aws_secret_access_key (str, optional): AWS secret access key for
            connecting the boto3 client. If not provided, will be loaded from
            your environment (via either the `AWS_SECRET_ACCESS_KEY`
            environment variable, or the `~/.aws/config` file).
            See [the boto3 credentials docs][1] for more information.
        - aws_session_token (str, optional): AWS session key for connecting the
            boto3 client. If not provided, will be loaded from your environment
            (via either the `AWS_SESSION_TOKEN` environment variable, or the
            `~/.aws/config` file). See [the boto3 credentials docs][1] for more
            information.
        - region_name (str, optional): AWS region name to launch ECS tasks in.
            If not provided, will be loaded from your environment (via either
            the `AWS_DEFAULT_REGION` environment variable, or the
            `~/.aws/config` file). See [the boto3 configuration docs][2] for
            more information.
        - cluster (str, optional): The AWS cluster to use, defaults to
            `"default"` if not provided.
        - launch_type (str, optional): The launch type to use, either
            `"FARGATE"` (default) or `"EC2"`.
        - task_role_arn (str, optional): The default task role ARN to use when
            registering ECS tasks created by this agent.
        - execution_role_arn (str, optional): The default execution role ARN
            to use when registering ECS tasks created by this agent.
        - botocore_config (dict, optional): Additional botocore configuration
            options to be passed to the boto3 client. See [the boto3
            configuration docs][2] for more information.


    [1]: https://boto3.amazonaws.com/v1/documentation/api/latest/guide/credentials.html

    [2]: https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html
    """

    def __init__(  # type: ignore
        self,
        agent_config_id: str = None,
        name: str = None,
        labels: Iterable[str] = None,
        env_vars: dict = None,
        max_polls: int = None,
        agent_address: str = None,
        no_cloud_logs: bool = False,
        task_definition_path: str = None,
        run_task_kwargs_path: str = None,
        aws_access_key_id: str = None,
        aws_secret_access_key: str = None,
        aws_session_token: str = None,
        region_name: str = None,
        cluster: str = None,
        launch_type: str = None,
        task_role_arn: str = None,
        execution_role_arn: str = None,
        botocore_config: dict = None,
    ) -> None:
        super().__init__(
            agent_config_id=agent_config_id,
            name=name,
            labels=labels,
            env_vars=env_vars,
            max_polls=max_polls,
            agent_address=agent_address,
            no_cloud_logs=no_cloud_logs,
        )

        from botocore.config import Config
        from prefect.utilities.aws import get_boto_client

        self.cluster = cluster
        self.launch_type = launch_type.upper() if launch_type else "FARGATE"
        self.task_role_arn = task_role_arn
        self.execution_role_arn = execution_role_arn

        # Load boto configuration. We want to use the standard retry mode by
        # default (which isn't boto's default due to backwards compatibility).
        # The logic below lets the user override our default retry mode either
        # in `botocore_config` or in their aws config file.
        #
        # See https://boto3.amazonaws.com/v1/documentation/api/latest/guide/retries.html
        # for more info.
        boto_config = Config(**botocore_config or {})
        if not boto_config.retries and not os.environ.get("AWS_RETRY_MODE"):
            boto_config.retries = {"mode": "standard"}

        self.boto_kwargs = dict(
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token,
            region_name=region_name,
            config=boto_config,
        )  # type: Dict[str, Any]

        self.ecs_client = get_boto_client("ecs", **self.boto_kwargs)

        # Load default task definition
        if not task_definition_path:
            task_definition_path = DEFAULT_TASK_DEFINITION_PATH
        try:
            self.task_definition = yaml.safe_load(
                read_bytes_from_path(task_definition_path)
            )
        except Exception:
            self.logger.error(
                "Failed to load default task definition from %r",
                task_definition_path,
                exc_info=True,
            )
            raise

        # Load default run_task kwargs
        if run_task_kwargs_path:
            try:
                self.run_task_kwargs = yaml.safe_load(
                    read_bytes_from_path(run_task_kwargs_path)
                )
            except Exception:
                self.logger.error(
                    "Failed to load default `run_task` kwargs from %r",
                    run_task_kwargs_path,
                    exc_info=True,
                )
                raise
        else:
            self.run_task_kwargs = {}

        # If running on fargate, auto-configure `networkConfiguration` for the
        # user if they didn't configure it themselves.
        if self.launch_type == "FARGATE" and not self.run_task_kwargs.get(
            "networkConfiguration"
        ):
            self.run_task_kwargs[
                "networkConfiguration"
            ] = self.infer_network_configuration()

    def infer_network_configuration(self) -> dict:
        """Infer default values for `networkConfiguration`.

        This is called when running on `FARGATE` with no `networkConfiguration`
        specified in the default `run_task_kwargs`. This makes it easier to get
        setup, as usually the default is what you want.

        Returns:
            - dict: Inferred `networkConfiguration`
        """
        from prefect.utilities.aws import get_boto_client

        self.logger.debug("Inferring default `networkConfiguration`...")

        ec2 = get_boto_client("ec2", **self.boto_kwargs)
        vpcs = ec2.describe_vpcs(Filters=[{"Name": "isDefault", "Values": ["true"]}])[
            "Vpcs"
        ]
        if vpcs:
            vpc_id = vpcs[0]["VpcId"]
            subnets = ec2.describe_subnets(
                Filters=[{"Name": "vpc-id", "Values": [vpc_id]}]
            )["Subnets"]
            if subnets:
                config = {
                    "awsvpcConfiguration": {
                        "subnets": [s["SubnetId"] for s in subnets],
                        "assignPublicIp": "ENABLED",
                    }
                }
                self.logger.debug("Using networkConfiguration=%r", config)
                return config

        msg = (
            "Failed to infer default networkConfiguration, please explicitly "
            "configure using `--run-task-kwargs`"
        )
        self.logger.error(msg)
        raise ValueError(msg)

    def deploy_flow(self, flow_run: GraphQLResult) -> str:
        """
        Deploy a flow run as an ECS task.

        Args:
            - flow_run (GraphQLResult): A GraphQLResult flow run object

        Returns:
            - str: Information about the deployment
        """
        self.logger.info("Deploying flow run %r", flow_run.id)

        run_config = self._get_run_config(flow_run, ECSRun)
        assert isinstance(run_config, ECSRun)  # mypy

        if run_config.task_definition_arn is None:
            # Register a new task definition
            self.logger.debug(
                "Registering new task definition for flow %s", flow_run.flow.id
            )
            taskdef = self.generate_task_definition(flow_run, run_config)
            resp = self.ecs_client.register_task_definition(**taskdef)
            taskdef_arn = resp["taskDefinition"]["taskDefinitionArn"]
            new_taskdef_arn = True
            self.logger.debug(
                "Registered task definition %s for flow %s",
                taskdef_arn,
                flow_run.flow.id,
            )
        else:
            from prefect.serialization.storage import StorageSchema
            from prefect.storage import Docker

            if isinstance(StorageSchema().load(flow_run.flow.storage), Docker):
                raise ValueError(
                    "Cannot provide `task_definition_arn` when using `Docker` storage"
                )
            taskdef_arn = run_config.task_definition_arn
            new_taskdef_arn = False
            self.logger.debug(
                "Using task definition %s for flow %s", taskdef_arn, flow_run.flow.id
            )

        # Get kwargs to pass to run_task
        kwargs = self.get_run_task_kwargs(flow_run, run_config)

        resp = self.ecs_client.run_task(taskDefinition=taskdef_arn, **kwargs)

        # Always deregister the task definition if a new one was registered
        if new_taskdef_arn:
            self.logger.debug("Deregistering task definition %s", taskdef_arn)
            self.ecs_client.deregister_task_definition(taskDefinition=taskdef_arn)

        if resp.get("tasks"):
            task_arn = resp["tasks"][0]["taskArn"]
            self.logger.debug("Started task %r for flow run %r", task_arn, flow_run.id)
            return f"Task {task_arn}"

        raise ValueError(
            "Failed to start task for flow run {0}. Failures: {1}".format(
                flow_run.id, resp.get("failures")
            )
        )

    def generate_task_definition(
        self, flow_run: GraphQLResult, run_config: ECSRun
    ) -> Dict[str, Any]:
        """Generate an ECS task definition from a flow run

        Args:
            - flow_run (GraphQLResult): A flow run object
            - run_config (ECSRun): The flow's run config

        Returns:
            - dict: a dictionary representation of an ECS task definition
        """
        if run_config.task_definition:
            taskdef = deepcopy(run_config.task_definition)
        elif run_config.task_definition_path:
            self.logger.debug(
                "Loading task definition template from %r",
                run_config.task_definition_path,
            )
            template_bytes = read_bytes_from_path(run_config.task_definition_path)
            taskdef = yaml.safe_load(template_bytes)
        else:
            taskdef = deepcopy(self.task_definition)

        slug = slugify.slugify(
            flow_run.flow.name,
            max_length=255 - len("prefect-"),
            word_boundary=True,
            save_order=True,
        )
        taskdef["family"] = f"prefect-{slug}"

        # Add some metadata tags for easier tracking by users
        taskdef.setdefault("tags", []).extend(
            [
                {"key": "prefect:flow-id", "value": flow_run.flow.id},
                {"key": "prefect:flow-version", "value": str(flow_run.flow.version)},
            ]
        )

        # Get the flow container (creating one if it doesn't already exist)
        containers = taskdef.setdefault("containerDefinitions", [])
        for container in containers:
            if container.get("name") == "flow":
                break
        else:
            container = {"name": "flow"}
            containers.append(container)

        # Set flow image
        container["image"] = image = get_flow_image(
            flow_run, default=container.get("image")
        )

        # Add `PREFECT__CONTEXT__IMAGE` environment variable
        env = {"PREFECT__CONTEXT__IMAGE": image}
        container_env = [{"name": k, "value": v} for k, v in env.items()]
        for entry in container.get("environment", []):
            if entry["name"] not in env:
                container_env.append(entry)
        container["environment"] = container_env

        # Ensure that cpu/memory are strings not integers
        if "cpu" in taskdef:
            taskdef["cpu"] = str(taskdef["cpu"])
        if "memory" in taskdef:
            taskdef["memory"] = str(taskdef["memory"])

        # Set executionRoleArn if configured
        # Note that we set this in both the task definition and the
        # run_task_kwargs since ECS requires this in the definition for FARGATE
        # tasks, but we also want to still configure it for users that provide
        # their own `task_definition_arn`.
        if run_config.execution_role_arn:
            taskdef["executionRoleArn"] = run_config.execution_role_arn
        elif self.execution_role_arn:
            taskdef["executionRoleArn"] = self.execution_role_arn

        # Set requiresCompatibilities if not already set
        if "requiresCompatibilities" not in taskdef:
            taskdef["requiresCompatibilities"] = [self.launch_type]

        return taskdef

    def get_run_task_kwargs(
        self, flow_run: GraphQLResult, run_config: ECSRun
    ) -> Dict[str, Any]:
        """Generate kwargs to pass to `ECS.client.run_task` for a flow run

        Args:
            - flow_run (GraphQLResult): A flow run object
            - run_config (ECSRun): The flow's run config

        Returns:
            - dict: kwargs to pass to `ECS.client.run_task`
        """
        # Set agent defaults
        out = deepcopy(self.run_task_kwargs)
        out["launchType"] = self.launch_type
        if self.cluster:
            out["cluster"] = self.cluster

        # Apply run-config kwargs, if any
        if run_config.run_task_kwargs:
            out = merge_run_task_kwargs(out, run_config.run_task_kwargs)

        # Find or create the flow container overrides
        overrides = out.setdefault("overrides", {})
        container_overrides = overrides.setdefault("containerOverrides", [])
        for container in container_overrides:
            if container.get("name") == "flow":
                break
        else:
            container = {"name": "flow"}
            container_overrides.append(container)

        # Set taskRoleArn if configured
        if run_config.task_role_arn:
            overrides["taskRoleArn"] = run_config.task_role_arn
        elif self.task_role_arn:
            overrides["taskRoleArn"] = self.task_role_arn

        # Set executionRoleArn if configured
        if run_config.execution_role_arn:
            overrides["executionRoleArn"] = run_config.execution_role_arn
        elif self.execution_role_arn:
            overrides["executionRoleArn"] = self.execution_role_arn

        # Set resource requirements, if provided
        # Also ensure that cpu/memory are strings not integers
        if run_config.cpu:
            overrides["cpu"] = str(run_config.cpu)
        elif "cpu" in overrides:
            overrides["cpu"] = str(overrides["cpu"])
        if run_config.memory:
            overrides["memory"] = str(run_config.memory)
        elif "memory" in overrides:
            overrides["memory"] = str(overrides["memory"])

        # Set flow run command
        container["command"] = ["/bin/sh", "-c", get_flow_run_command(flow_run)]

        # Add `PREFECT__LOGGING__LEVEL` environment variable
        env = {"PREFECT__LOGGING__LEVEL": config.logging.level}

        # Populate environment variables from the following sources,
        # with precedence:
        # - Values required for flow execution, hardcoded below
        # - Values set on the ECSRun object
        # - Values set using the `--env` CLI flag on the agent
        env.update(self.env_vars)
        if run_config.env:
            env.update(run_config.env)
        env.update(
            {
                "PREFECT__CLOUD__USE_LOCAL_SECRETS": "false",
                "PREFECT__ENGINE__FLOW_RUNNER__DEFAULT_CLASS": "prefect.engine.cloud.CloudFlowRunner",
                "PREFECT__ENGINE__TASK_RUNNER__DEFAULT_CLASS": "prefect.engine.cloud.CloudTaskRunner",
                "PREFECT__BACKEND": config.backend,
                "PREFECT__CLOUD__API": config.cloud.api,
                "PREFECT__CONTEXT__FLOW_RUN_ID": flow_run.id,
                "PREFECT__CONTEXT__FLOW_ID": flow_run.flow.id,
                "PREFECT__LOGGING__LOG_TO_CLOUD": str(self.log_to_cloud).lower(),
                "PREFECT__CLOUD__AUTH_TOKEN": config.cloud.agent.auth_token,
                "PREFECT__CLOUD__AGENT__LABELS": str(self.labels),
            }
        )
        container_env = [{"name": k, "value": v} for k, v in env.items()]
        for entry in container.get("environment", []):
            if entry["name"] not in env:
                container_env.append(entry)
        container["environment"] = container_env

        return out



import io
from contextlib import closing
from typing import TYPE_CHECKING, Any

import pendulum
from slugify import slugify

import prefect
from prefect.engine.results import S3Result
from prefect.storage import Storage
from prefect.utilities.storage import (
    extract_flow_from_file,
    flow_from_bytes_pickle,
    flow_to_bytes_pickle,
)

if TYPE_CHECKING:
    from prefect.core.flow import Flow


class S3(Storage):
    """
    S3 storage class.  This class represents the Storage interface for Flows stored as
    bytes in an S3 bucket.

    This storage class optionally takes a `key` which will be the name of the Flow object
    when stored in S3. If this key is not provided the Flow upload name will take the form
    `slugified-flow-name/slugified-current-timestamp`.

    Args:
        - bucket (str): the name of the S3 Bucket to store Flows
        - key (str, optional): a unique key to use for uploading a Flow to S3. This
            is only useful when storing a single Flow using this storage object.
        - stored_as_script (bool, optional): boolean for specifying if the flow has been stored
            as a `.py` file. Defaults to `False`
        - local_script_path (str, optional): the path to a local script to upload when `stored_as_script`
            is set to `True`. If not set then the value of `local_script_path` from `prefect.context` is
            used. If neither are set then script will not be uploaded and users should manually place the
            script file in the desired `key` location in an S3 bucket.
        - client_options (dict, optional): Additional options for the `boto3` client.
        - **kwargs (Any, optional): any additional `Storage` initialization options
    """

    def __init__(
        self,
        bucket: str,
        key: str = None,
        stored_as_script: bool = False,
        local_script_path: str = None,
        client_options: dict = None,
        **kwargs: Any,
    ) -> None:
        self.bucket = bucket
        self.key = key
        self.local_script_path = local_script_path or prefect.context.get(
            "local_script_path", None
        )

        self.client_options = client_options

        result = S3Result(bucket=bucket, boto3_kwargs=client_options)
        super().__init__(
            result=result,
            stored_as_script=stored_as_script,
            **kwargs,
        )

    def get_flow(self, flow_name: str) -> "Flow":
        """
        Given a flow name within this Storage object, load and return the Flow.

        Args:
            - flow_name (str): the name of the flow to return.

        Returns:
            - Flow: the requested flow
        """
        if flow_name not in self.flows:
            raise ValueError("Flow is not contained in this Storage")
        key = self.flows[flow_name]

        self.logger.info(f"Downloading flow from s3://{self.bucket}/{key}")

        try:
            obj = self._boto3_client.get_object(Bucket=self.bucket, Key=key)
            body = obj["Body"]
            with closing(body):
                output = body.read()
        except Exception as err:
            self.logger.error("Error downloading Flow from S3: {}".format(err))
            raise

        self.logger.info(
            "Flow successfully downloaded. ETag: %s, LastModified: %s, VersionId: %s",
            obj["ETag"],
            obj["LastModified"].isoformat(),
            obj.get("VersionId"),
        )

        if self.stored_as_script:
            return extract_flow_from_file(file_contents=output, flow_name=flow_name)  # type: ignore

        return flow_from_bytes_pickle(output)

    def add_flow(self, flow: "Flow") -> str:
        """
        Method for storing a new flow as bytes in the local filesytem.

        Args:
            - flow (Flow): a Prefect Flow to add

        Returns:
            - str: the location of the newly added flow in this Storage object

        Raises:
            - ValueError: if a flow with the same name is already contained in this storage
        """
        if flow.name in self:
            raise ValueError(
                'Name conflict: Flow with the name "{}" is already present in this storage.'.format(
                    flow.name
                )
            )

        # Create key for Flow that uniquely identifies Flow object in S3
        key = self.key or "{}/{}".format(
            slugify(flow.name), slugify(pendulum.now("utc").isoformat())
        )

        # Append .py extension when storing as script
        # if self.stored_as_script and not key.endswith(".py"):
        #     key = key + ".py"

        self.flows[flow.name] = key
        self._flows[flow.name] = flow
        return key

    def build(self) -> "Storage":
        """
        Build the S3 storage object by uploading Flows to an S3 bucket. This will upload
        all of the flows found in `storage.flows`. If there is an issue uploading to the
        S3 bucket an error will be logged.

        Returns:
            - Storage: an S3 object that contains information about how and where
                each flow is stored

        Raises:
            - botocore.ClientError: if there is an issue uploading a Flow to S3
        """
        self.run_basic_healthchecks()

        from botocore.exceptions import ClientError

        if self.stored_as_script:
            if self.local_script_path:
                for flow_name, flow in self._flows.items():
                    self.logger.info(
                        "Uploading script {} to {} in {}".format(
                            self.local_script_path, self.flows[flow.name], self.bucket
                        )
                    )

                    try:
                        self._boto3_client.upload_file(
                            self.local_script_path, self.bucket, self.flows[flow_name]
                        )
                    except ClientError as err:
                        self.logger.error(
                            "Error uploading Flow script to S3 bucket {}: {}".format(
                                self.bucket, err
                            )
                        )
                        raise err
            else:
                if not self.key:
                    raise ValueError(
                        "A `key` must be provided to show where flow `.py` file is stored in S3 or "
                        "provide a `local_script_path` pointing to a local script that contains the "
                        "flow."
                    )
            return self

        for flow_name, flow in self._flows.items():
            # Pickle Flow
            data = flow_to_bytes_pickle(flow)

            # Write pickled Flow to stream
            stream = io.BytesIO(data)

            self.logger.info(
                "Uploading {} to {}".format(self.flows[flow_name], self.bucket)
            )

            try:
                self._boto3_client.upload_fileobj(
                    stream, Bucket=self.bucket, Key=self.flows[flow_name]
                )
            except ClientError as err:
                self.logger.error(
                    "Error uploading Flow to S3 bucket {}: {}".format(self.bucket, err)
                )
                raise err

        return self

    @property
    def _boto3_client(self):  # type: ignore
        from prefect.utilities.aws import get_boto_client

        kwargs = self.client_options or {}
        return get_boto_client(resource="s3", credentials=None, **kwargs)



import os
from typing import TYPE_CHECKING, Any

import prefect
from prefect.client import Secret
from prefect.storage import Storage
from prefect.utilities.storage import extract_flow_from_file

if TYPE_CHECKING:
    from prefect.core.flow import Flow
    from github import Github


class GitHub(Storage):
    """
    GitHub storage class. This class represents the Storage interface for Flows stored
    in `.py` files in a GitHub repository.

    This class represents a mapping of flow name to file paths contained in the git repo,
    meaning that all flow files should be pushed independently. A typical workflow using
    this storage type might look like the following:

    - Compose flow `.py` file where flow has GitHub storage:

    ```python
    flow = Flow("my-flow")
    flow.storage = GitHub(repo="my/repo", path="/flows/flow.py")
    ```

    - Push this `flow.py` file to the `my/repo` repository under `/flows/flow.py`.

    - Call `prefect register -f flow.py` to register this flow with GitHub storage.

    Args:
        - repo (str): the name of a GitHub repository to store this Flow
        - path (str): a path pointing to a flow file in the repo
        - ref (str, optional): a commit SHA-1 value, tag, or branch name. If not specified,
            defaults to the default branch for the repo.
        - access_token_secret (str, optional): The name of a Prefect secret
            that contains a GitHub access token to use when loading flows from
            this storage.
        - base_url(str, optional): the Github REST api url for the repo. If not specified,
            https://api.github.com is used.
        - **kwargs (Any, optional): any additional `Storage` initialization options
    """

    def __init__(
        self,
        repo: str,
        path: str,
        ref: str = None,
        access_token_secret: str = None,
        base_url: str = None,
        **kwargs: Any,
    ) -> None:
        self.repo = repo
        self.path = path
        self.ref = ref
        self.access_token_secret = access_token_secret
        self.base_url = base_url

        super().__init__(**kwargs)

    def get_flow(self, flow_name: str) -> "Flow":
        """
        Given a flow name within this Storage object, load and return the Flow.

        Args:
            - flow_name (str): the name of the flow to return.

        Returns:
            - Flow: the requested flow
        """
        try:
            from github import UnknownObjectException
        except ImportError as exc:
            raise ImportError(
                "Unable to import Github, please ensure you have installed the github extra"
            ) from exc

        if flow_name not in self.flows:
            raise ValueError("Flow is not contained in this Storage")
        path = self.flows[flow_name]

        # Log info about the active storage object. Only include `ref` if
        # explicitly set.
        self.logger.info(
            "Downloading flow from GitHub storage - repo: %r, path: %r%s",
            self.repo,
            path,
            f", ref: {self.ref!r}" if self.ref is not None else "",
        )

        client = self._get_github_client()

        try:
            repo = client.get_repo(self.repo)
        except UnknownObjectException:
            self.logger.error(
                "Repo %r not found. Check that it exists (and is spelled correctly), "
                "and that you have configured the proper credentials for accessing it.",
                self.repo,
            )
            raise

        # Use the default branch if unspecified
        ref = self.ref or repo.default_branch

        # Get the current commit sha for this ref
        try:
            commit = repo.get_commit(ref).sha
        except UnknownObjectException:
            self.logger.error("Ref %r not found in repo %r.", ref, self.repo)
            raise

        try:
            contents = repo.get_contents(path, ref=commit)
            assert not isinstance(contents, list)  # mypy
            decoded_contents = contents.decoded_content.decode()
        except UnknownObjectException:
            self.logger.error(
                "File %r not found in repo %r, ref %r", path, self.repo, ref
            )
            raise

        self.logger.info("Flow successfully downloaded. Using commit: %s", commit)

        return extract_flow_from_file(
            file_contents=decoded_contents, flow_name=flow_name
        )

    def add_flow(self, flow: "Flow") -> str:
        """
        Method for storing a new flow as bytes in the local filesytem.

        Args:
            - flow (Flow): a Prefect Flow to add

        Returns:
            - str: the location of the added flow in the repo

        Raises:
            - ValueError: if a flow with the same name is already contained in this storage
        """
        if flow.name in self:
            raise ValueError(
                'Name conflict: Flow with the name "{}" is already present in this storage.'.format(
                    flow.name
                )
            )

        self.flows[flow.name] = self.path  # type: ignore
        self._flows[flow.name] = flow
        return self.path  # type: ignore

    def _get_github_client(self) -> "Github":
        from github import Github

        if self.access_token_secret is not None:
            # If access token secret specified, load it
            access_token = Secret(self.access_token_secret).get()
        else:
            # Otherwise, fallback to loading from local secret or environment variable
            access_token = prefect.context.get("secrets", {}).get("GITHUB_ACCESS_TOKEN")
            if access_token is None:
                access_token = os.getenv("GITHUB_ACCESS_TOKEN")

        if self.base_url:
            return Github(login_or_token=access_token, base_url=self.base_url)
        else:
            return Github(access_token)



from tempfile import TemporaryDirectory
from typing import Any


class TemporaryGitRepo:
    """
    Temporary cloning and interacting with a git repository

    Args:
        - git_clone_url (str): Url to git clone
        - branch_name (str, optional): branch name, if not specified and `tag` not specified,
            repo default branch latest commit will be used
        - tag (str, optional): tag name, if not specified and `branch_name` not specified,
            repo default branch latest commit will be used
        - commit (str, optional): a commit SHA-1 value, if not specified and `branch_name`
            and `tag` not specified, repo default branch latest commit will be used
        - clone_depth (int): the number of history revisions in cloning, defaults to 1
    """

    def __init__(
        self,
        git_clone_url: str,
        branch_name: str = None,
        tag: str = None,
        commit: str = None,
        clone_depth: int = 1,
    ) -> None:
        if tag and branch_name:
            raise ValueError(
                "Either `tag` or `branch_name` can be specified, but not both"
            )
        self.git_clone_url = git_clone_url
        self.branch_name = branch_name
        self.tag = tag
        self.commit = commit
        self.clone_depth = clone_depth

    def __enter__(self) -> "TemporaryGitRepo":
        try:
            from dulwich.porcelain import clone
        except ImportError as exc:
            raise ImportError(
                "Unable to import dulwich, please ensure you have installed the git extra"
            ) from exc

        self.temp_dir = TemporaryDirectory()
        self.repo = clone(
            source=self.git_clone_url, target=self.temp_dir.name, depth=self.clone_depth
        )
        if self.branch_name is not None or self.tag is not None:
            self.checkout_ref()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self.temp_dir.cleanup()

    def checkout_ref(self) -> None:
        """
        Checkout a specific ref from the repo
        """
        try:
            from dulwich.index import build_index_from_tree
        except ImportError as exc:
            raise ImportError(
                "Unable to import dulwich, please ensure you have installed the git extra"
            ) from exc

        build_index_from_tree(
            self.repo.path,
            self.repo.index_path(),
            self.repo.object_store,
            self.get_tree_id_for_branch_or_tag(),
        )

    @property
    def branch_or_tag_ref(self) -> bytes:
        """
        Get the branch or tag ref for the current repo
        """
        if self.branch_name is not None:
            return f"refs/remotes/origin/{self.branch_name}".encode("utf-8")
        elif self.tag is not None:
            return f"refs/tags/{self.tag}".encode("utf-8")
        raise ValueError(
            "Either `tag` or `branch_name` must be specified to get a tree id"
        )

    def get_tree_id_for_branch_or_tag(self) -> str:
        """
        Gets the tree id for relevant branch or tag
        """
        if self.commit is not None:
            return self.repo[self.commit.encode("utf-8")]
        return self.repo[self.repo.get_refs()[self.branch_or_tag_ref]].tree



from collections import defaultdict
from datetime import datetime
import os
import pytz
import time
import uuid
from typing import Optional, Iterable, List, Any

import json
import yaml

import prefect
from prefect import config
from prefect.agent import Agent
from prefect.engine.state import Failed
from prefect.run_configs import KubernetesRun
from prefect.utilities.agent import get_flow_image, get_flow_run_command
from prefect.utilities.exceptions import ClientError
from prefect.utilities.filesystems import read_bytes_from_path
from prefect.utilities.graphql import GraphQLResult

DEFAULT_JOB_TEMPLATE_PATH = os.path.join(os.path.dirname(__file__), "job_template.yaml")


def _get_or_create(d: dict, key: str, val: Any = None) -> Any:
    """Get a (possibly nested) field from a dict, creating intermediate values
    if needed."""
    if val is None:
        val = {}
    path = key.split(".")
    for k in path[:-1]:
        d = d.setdefault(k, {})
    return d.setdefault(path[-1], val)


class KubernetesAgent(Agent):
    """
    Agent which deploys flow runs as Kubernetes jobs. Currently this is required to either
    run on a k8s cluster or on a local machine where the kube_config is pointing at the
    desired cluster. Information on using the Kubernetes Agent can be found at
    https://docs.prefect.io/orchestration/agents/kubernetes.html

    Args:
        - agent_config_id (str, optional): An optional agent configuration ID that can be used to set
            configuration based on an agent from a backend API. If set all configuration values will be
            pulled from backend agent configuration.
        - namespace (str, optional): A Kubernetes namespace to create jobs in. Defaults
            to the environment variable `NAMESPACE` or `default`.
        - service_account_name (str, optional): A kubernetes service account name to use by
            default for created jobs. May be overridden per-flow by specifying
            on a flow's `KubernetesRun` run config.
        - image_pull_secrets (list, optional): A list of image pull secrets to use by default
            for created jobs. May be overridden per-flow by specifying on a flow's
            `KubernetesRun` run config.
        - job_template_path (str, optional): A path to a job template file to use instead
            of the default.
        - name (str, optional): An optional name to give this agent. Can also be set through
            the environment variable `PREFECT__CLOUD__AGENT__NAME`. Defaults to "agent"
        - labels (List[str], optional): a list of labels, which are arbitrary string
            identifiers used by Prefect Agents when polling for work
        - env_vars (dict, optional): a dictionary of environment variables and values that will
            be set on each flow run that this agent submits for execution and in the agent's
            own environment
        - max_polls (int, optional): maximum number of times the agent will poll Prefect Cloud
            for flow runs; defaults to infinite
        - agent_address (str, optional):  Address to serve internal API at. Currently this is
            just health checks for use by an orchestration layer. Leave blank for no API server
            (default).
        - no_cloud_logs (bool, optional): Disable logging to a Prefect backend for this agent
            and all deployed flow runs
        - volume_mounts (list, optional): A list of volumeMounts to mount when a job is
            run. The volumeMounts in the list should be specified as dicts
            i.e `[{"name": "my-vol", "mountPath": "/mnt/my-mount"}]`
        - volumes (list, optional): A list of volumes to make available to be mounted when a
            job is run. The volumes in the list should be specified as nested dicts.
            i.e `[{"name": "my-vol", "csi": {"driver": "secrets-store.csi.k8s.io"}}]`
        - delete_finished_jobs (bool, optional): A boolean to toggle if finished Prefect jobs
            in the agent's namespace should be deleted. Defaults to the environment variable
            `DELETE_FINISHED_JOBS` or `True`.
    """

    def __init__(
        self,
        agent_config_id: str = None,
        namespace: str = None,
        service_account_name: str = None,
        image_pull_secrets: Iterable[str] = None,
        job_template_path: str = None,
        name: str = None,
        labels: Iterable[str] = None,
        env_vars: dict = None,
        max_polls: int = None,
        agent_address: str = None,
        no_cloud_logs: bool = False,
        volume_mounts: List[dict] = None,
        volumes: List[dict] = None,
        delete_finished_jobs: bool = True,
    ) -> None:
        super().__init__(
            agent_config_id=agent_config_id,
            name=name,
            labels=labels,
            env_vars=env_vars,
            max_polls=max_polls,
            agent_address=agent_address,
            no_cloud_logs=no_cloud_logs,
        )

        self.namespace = namespace or os.getenv("NAMESPACE", "default")
        self.service_account_name = service_account_name or os.getenv(
            "SERVICE_ACCOUNT_NAME"
        )
        if image_pull_secrets is None:
            image_pull_secrets_env = os.getenv("IMAGE_PULL_SECRETS")
            image_pull_secrets = (
                [s.strip() for s in image_pull_secrets_env.split(",")]
                if image_pull_secrets_env is not None
                else None
            )
        self.image_pull_secrets = image_pull_secrets
        self.job_template_path = job_template_path or DEFAULT_JOB_TEMPLATE_PATH
        self.volume_mounts = volume_mounts
        self.volumes = volumes
        self.delete_finished_jobs = delete_finished_jobs and (
            os.getenv("DELETE_FINISHED_JOBS", "True") == "True"
        )

        from kubernetes import client, config

        try:
            self.logger.debug("Loading incluster configuration")
            config.load_incluster_config()
        except config.config_exception.ConfigException as exc:
            self.logger.warning(
                "{} Using out of cluster configuration option.".format(exc)
            )
            self.logger.debug("Loading out of cluster configuration")
            config.load_kube_config()

        self.batch_client = client.BatchV1Api()
        self.core_client = client.CoreV1Api()
        self.k8s_client = client

        min_datetime = datetime.min.replace(tzinfo=pytz.UTC)
        self.job_pod_event_timestamps = defaultdict(  # type: ignore
            lambda: defaultdict(lambda: min_datetime)
        )

        self.logger.debug(f"Namespace: {self.namespace}")

    def manage_jobs(self) -> None:
        """
        This function checks if jobs are `Failed` or `Succeeded` and if they are then the jobs are
        deleted from the namespace. If one of the job's pods happen to run into image pulling errors
        then the flow run is failed and the job is still deleted.
        """
        self.logger.debug(
            "Retrieving information of jobs that are currently in the cluster..."
        )

        more = True
        _continue = ""
        while more:
            try:
                jobs = self.batch_client.list_namespaced_job(
                    namespace=self.namespace,
                    label_selector="prefect.io/identifier",
                    limit=20,
                    _continue=_continue,
                )
                _continue = jobs.metadata._continue
                more = bool(_continue)

                for job in jobs.items:
                    delete_job = job.status.failed or job.status.succeeded
                    job_name = job.metadata.name
                    flow_run_id = job.metadata.labels.get("prefect.io/flow_run_id")

                    # Check for pods that are stuck with image pull errors
                    if not delete_job:
                        pods = self.core_client.list_namespaced_pod(
                            namespace=self.namespace,
                            label_selector="prefect.io/identifier={}".format(
                                job.metadata.labels.get("prefect.io/identifier")
                            ),
                        )

                        for pod in pods.items:
                            pod_name = pod.metadata.name
                            if pod.status.container_statuses:
                                for container_status in pod.status.container_statuses:
                                    waiting = container_status.state.waiting
                                    if waiting and (
                                        waiting.reason == "ErrImagePull"
                                        or waiting.reason == "ImagePullBackOff"
                                    ):
                                        self.logger.debug(
                                            f"Failing flow run {flow_run_id} due to pod {waiting.reason}"
                                        )
                                        try:
                                            self.client.set_flow_run_state(
                                                flow_run_id=flow_run_id,
                                                state=Failed(
                                                    message="Kubernetes Error: {}".format(
                                                        container_status.state.waiting.message
                                                    )
                                                ),
                                            )
                                        except ClientError as exc:
                                            self.logger.error(
                                                "Error attempting to set flow run state for "
                                                f"{flow_run_id}: "
                                                f"{exc}"
                                            )

                                        delete_job = True
                                        break

                            # Report recent events for pending pods to flow run logs
                            if pod.status.phase == "Pending":
                                pod_events = self.core_client.list_namespaced_event(
                                    namespace=self.namespace,
                                    field_selector="involvedObject.name={}".format(
                                        pod_name
                                    ),
                                    timeout_seconds=30,
                                )

                                for event in sorted(
                                    pod_events.items, key=lambda x: x.last_timestamp
                                ):
                                    # Skip old events
                                    if (
                                        event.last_timestamp
                                        < self.job_pod_event_timestamps[job_name][
                                            pod_name
                                        ]
                                    ):
                                        continue

                                    self.job_pod_event_timestamps[job_name][
                                        pod_name
                                    ] = event.last_timestamp

                                    log_msg = (
                                        f"Event: {event.reason!r} on pod {pod_name!r}\n"
                                        f"\tMessage: {event.message}"
                                    )

                                    # Send pod failure information to flow run logs
                                    self.client.write_run_logs(
                                        [
                                            dict(
                                                flow_run_id=flow_run_id,
                                                name="k8s-infra",
                                                message=log_msg,
                                                level="DEBUG",
                                                timestamp=event.last_timestamp.isoformat(),
                                            )
                                        ]
                                    )

                    # Report failed pods
                    if job.status.failed:
                        pods = self.core_client.list_namespaced_pod(
                            namespace=self.namespace,
                            label_selector="prefect.io/identifier={}".format(
                                job.metadata.labels.get("prefect.io/identifier")
                            ),
                        )

                        failed_pods = []
                        for pod in pods.items:
                            if pod.status.phase != "Failed":
                                continue

                            # Format pod failure error message
                            failed_pods.append(pod.metadata.name)
                            pod_status_logs = [f"Pod {pod.metadata.name} failed."]
                            if not pod.status.container_statuses:
                                pod_status_logs.append(
                                    "\tNo container statuses found for pod"
                                )
                            else:
                                for status in pod.status.container_statuses:
                                    state = (
                                        "running"
                                        if status.state.running
                                        else "waiting"
                                        if status.state.waiting
                                        else "terminated"
                                        if status.state.terminated
                                        else "Not Found"
                                    )
                                    pod_status_logs.append(
                                        f"\tContainer '{status.name}' state: {state}"
                                    )

                                    if status.state.terminated:
                                        pod_status_logs.append(
                                            f"\t\tExit Code:: {status.state.terminated.exit_code}"
                                        )
                                        if status.state.terminated.message:
                                            pod_status_logs.append(
                                                f"\t\tMessage: {status.state.terminated.message}"
                                            )
                                        if status.state.terminated.reason:
                                            pod_status_logs.append(
                                                f"\t\tReason: {status.state.terminated.reason}"
                                            )
                                        if status.state.terminated.signal:
                                            pod_status_logs.append(
                                                f"\t\tSignal: {status.state.terminated.signal}"
                                            )

                            # Send pod failure information to flow run logs
                            self.client.write_run_logs(
                                [
                                    dict(
                                        flow_run_id=flow_run_id,
                                        name="k8s-infra",
                                        message="\n".join(pod_status_logs),
                                        level="ERROR",
                                    )
                                ]
                            )

                        # If there are failed pods and the run is not finished, fail the run
                        if (
                            failed_pods
                            and not self.client.get_flow_run_state(
                                flow_run_id
                            ).is_finished()
                        ):
                            self.logger.debug(
                                f"Failing flow run {flow_run_id} due to the failed pods {failed_pods}"
                            )
                            try:
                                self.client.set_flow_run_state(
                                    flow_run_id=flow_run_id,
                                    state=Failed(
                                        message="Kubernetes Error: pods {} failed for this job".format(
                                            failed_pods
                                        )
                                    ),
                                )
                            except ClientError as exc:
                                self.logger.error(
                                    f"Error attempting to set flow run state for {flow_run_id}: "
                                    f"{exc}"
                                )

                    # Delete job if it is successful or failed
                    if delete_job and self.delete_finished_jobs:
                        self.logger.debug(f"Deleting job {job_name}")
                        try:
                            self.job_pod_event_timestamps.pop(job_name, None)
                            self.batch_client.delete_namespaced_job(
                                name=job_name,
                                namespace=self.namespace,
                                body=self.k8s_client.V1DeleteOptions(
                                    propagation_policy="Foreground"
                                ),
                            )
                        except self.k8s_client.rest.ApiException as exc:
                            if exc.status != 404:
                                self.logger.error(
                                    f"{exc.status} error attempting to delete job {job_name}"
                                )
            except self.k8s_client.rest.ApiException as exc:
                if exc.status == 410:
                    self.logger.debug("Refreshing job listing token...")
                    _continue = ""
                    continue
                else:
                    self.logger.debug(exc)

    def heartbeat(self) -> None:
        """
        Check status of jobs created by this agent, delete completed jobs and failed containers.
        """
        try:
            self.manage_jobs()
        except Exception:
            self.logger.error("Error while managing existing k8s jobs", exc_info=True)
        super().heartbeat()

    def deploy_flow(self, flow_run: GraphQLResult) -> str:
        """
        Deploy flow runs on to a k8s cluster as jobs

        Args:
            - flow_run (GraphQLResult): A GraphQLResult flow run object

        Returns:
            - str: Information about the deployment
        """
        import urllib3.exceptions

        self.logger.info("Deploying flow run {}".format(flow_run.id))  # type: ignore

        job_spec = self.generate_job_spec(flow_run=flow_run)
        job_name = job_spec["metadata"]["name"]

        self.logger.debug("Creating namespaced job {}".format(job_name))
        attempts = 3
        while True:
            try:
                self.batch_client.create_namespaced_job(
                    namespace=self.namespace, body=job_spec
                )
                break
            except self.k8s_client.rest.ApiException as exc:
                if exc.status == 409:
                    # object already exists, previous submission was successful
                    # even though it errored
                    break
                raise
            except urllib3.exceptions.HTTPError:
                attempts -= 1
                if attempts == 0:
                    raise
                self.logger.warning(
                    "Error submitting job %s, retrying...", job_name, exc_info=True
                )
                time.sleep(1)

        self.logger.debug("Job {} created".format(job_name))

        return "Job {}".format(job_name)

    def generate_job_spec(self, flow_run: GraphQLResult) -> dict:
        """Generate a k8s job spec for a flow run

        Args:
            - flow_run (GraphQLResult): A flow run object

        Returns:
            - dict: a dictionary representation of a k8s job for flow execution
        """
        run_config = self._get_run_config(flow_run, KubernetesRun)
        assert run_config is None or isinstance(run_config, KubernetesRun)  # mypy
        if run_config is not None:
            return self.generate_job_spec_from_run_config(flow_run, run_config)
        else:
            return self.generate_job_spec_from_environment(flow_run)

    def generate_job_spec_from_environment(
        self, flow_run: GraphQLResult, image: str = None
    ) -> dict:
        """
        Populate a k8s job spec. This spec defines a k8s job that handles
        executing a flow. This method runs each time the agent receives
        a flow to run.

        That job spec can optionally be customized by setting the
        following environment variables on the agent.

        - `NAMESPACE`: the k8s namespace the job will run in, defaults to `"default"`
        - `JOB_MEM_REQUEST`: memory requested, for example, `256Mi` for 256 MB. If this
                environment variable is not set, the cluster's defaults will be used.
        - `JOB_MEM_LIMIT`: memory limit, for example, `512Mi` For 512 MB. If this
                environment variable is not set, the cluster's defaults will be used.
        - `JOB_CPU_REQUEST`: CPU requested, defaults to `"100m"`
        - `JOB_CPU_LIMIT`: CPU limit, defaults to `"100m"`
        - `IMAGE_PULL_POLICY`: policy for pulling images. Defaults to `"IfNotPresent"`.
        - `IMAGE_PULL_SECRETS`: name of an existing k8s secret that can be used to pull
                images. This is necessary if your flow uses an image that is in a non-public
                container registry, such as Amazon ECR, or in a public registry that requires
                authentication to avoid hitting rate limits. To specify multiple image pull
                secrets, provide a comma-delimited string with no spaces, like
                `"some-secret,other-secret"`.
        - `SERVICE_ACCOUNT_NAME`: name of a service account to run the job as.
                By default, none is specified.
        - `YAML_TEMPLATE`: a path to where the YAML template should be loaded from. defaults
                to the embedded `job_spec.yaml`.

        Args:
            - flow_run (GraphQLResult): A flow run object
            - image (str, optional): The full name of an image to use for the job

        Returns:
            - dict: a dictionary representation of a k8s job for flow execution
        """
        identifier = str(uuid.uuid4())[:8]
        yaml_path = os.getenv(
            "YAML_TEMPLATE", os.path.join(os.path.dirname(__file__), "job_spec.yaml")
        )
        with open(yaml_path, "r") as job_file:
            job = yaml.safe_load(job_file)

        job_name = "prefect-job-{}".format(identifier)

        # Populate job metadata for identification
        k8s_labels = {
            "prefect.io/identifier": identifier,
            "prefect.io/flow_run_id": flow_run.id,  # type: ignore
            "prefect.io/flow_id": flow_run.flow.id,  # type: ignore
        }
        job["metadata"]["name"] = job_name
        job["metadata"]["labels"].update(**k8s_labels)
        job["spec"]["template"]["metadata"]["labels"].update(**k8s_labels)

        # Use provided image for job
        if image is None:
            image = get_flow_image(flow_run=flow_run)
        job["spec"]["template"]["spec"]["containers"][0]["image"] = image

        self.logger.debug("Using image {} for job".format(image))

        # Datermine flow run command
        job["spec"]["template"]["spec"]["containers"][0]["args"] = [
            get_flow_run_command(flow_run)
        ]

        # Populate environment variables for flow run execution
        env = job["spec"]["template"]["spec"]["containers"][0]["env"]

        env[0]["value"] = config.cloud.api or "https://api.prefect.io"
        env[1]["value"] = config.cloud.agent.auth_token
        env[2]["value"] = flow_run.id  # type: ignore
        env[3]["value"] = flow_run.flow.id  # type: ignore
        env[4]["value"] = self.namespace
        env[5]["value"] = str(self.labels)
        env[6]["value"] = str(self.log_to_cloud).lower()
        env[7]["value"] = self.env_vars.get(
            "PREFECT__LOGGING__LEVEL", config.logging.level
        )

        # append all user provided values
        for key, value in self.env_vars.items():
            env.append(dict(name=key, value=value))

        # Use image pull secrets if provided
        if self.image_pull_secrets:
            for idx, secret_name in enumerate(self.image_pull_secrets):
                # this check preserves behavior from previous releases,
                # where prefect would only overwrite the first entry in
                # imagePullSecrets
                if idx == 0:
                    job["spec"]["template"]["spec"]["imagePullSecrets"][0] = {
                        "name": secret_name
                    }
                else:
                    job["spec"]["template"]["spec"]["imagePullSecrets"].append(
                        {"name": secret_name}
                    )
        else:
            del job["spec"]["template"]["spec"]["imagePullSecrets"]

        # Set resource requirements if provided
        resources = job["spec"]["template"]["spec"]["containers"][0]["resources"]
        if os.getenv("JOB_MEM_REQUEST"):
            resources["requests"]["memory"] = os.getenv("JOB_MEM_REQUEST")
        if os.getenv("JOB_MEM_LIMIT"):
            resources["limits"]["memory"] = os.getenv("JOB_MEM_LIMIT")
        if os.getenv("JOB_CPU_REQUEST"):
            resources["requests"]["cpu"] = os.getenv("JOB_CPU_REQUEST")
        if os.getenv("JOB_CPU_LIMIT"):
            resources["limits"]["cpu"] = os.getenv("JOB_CPU_LIMIT")
        if self.volume_mounts:
            job["spec"]["template"]["spec"]["containers"][0][
                "volumeMounts"
            ] = self.volume_mounts
        else:
            del job["spec"]["template"]["spec"]["containers"][0]["volumeMounts"]
        if self.volumes:
            job["spec"]["template"]["spec"]["volumes"] = self.volumes
        else:
            del job["spec"]["template"]["spec"]["volumes"]
        if os.getenv("IMAGE_PULL_POLICY"):
            job["spec"]["template"]["spec"]["containers"][0][
                "imagePullPolicy"
            ] = os.getenv("IMAGE_PULL_POLICY")
        if self.service_account_name:
            job["spec"]["template"]["spec"][
                "serviceAccountName"
            ] = self.service_account_name

        return job

    def generate_job_spec_from_run_config(
        self, flow_run: GraphQLResult, run_config: KubernetesRun
    ) -> dict:
        """Generate a k8s job spec for a flow run.

        Args:
            - flow_run (GraphQLResult): A flow run object
            - run_config (KubernetesRun): The flow run's run_config

        Returns:
            - dict: a dictionary representation of a k8s job for flow execution
        """
        if run_config.job_template:
            job = run_config.job_template
        else:
            job_template_path = run_config.job_template_path or self.job_template_path
            self.logger.debug("Loading job template from %r", job_template_path)
            template_bytes = read_bytes_from_path(job_template_path)
            job = yaml.safe_load(template_bytes)

        identifier = uuid.uuid4().hex[:8]

        job_name = f"prefect-job-{identifier}"

        # Populate job metadata for identification
        k8s_labels = {
            "prefect.io/identifier": identifier,
            "prefect.io/flow_run_id": flow_run.id,  # type: ignore
            "prefect.io/flow_id": flow_run.flow.id,  # type: ignore
        }
        _get_or_create(job, "metadata.labels")
        _get_or_create(job, "spec.template.metadata.labels")
        job["metadata"]["name"] = job_name
        job["metadata"]["labels"].update(**k8s_labels)
        job["spec"]["template"]["metadata"]["labels"].update(**k8s_labels)
        pod_spec = job["spec"]["template"]["spec"]

        # Configure `service_account_name` if specified
        if run_config.service_account_name is not None:
            # On run-config, always override
            service_account_name = (
                run_config.service_account_name
            )  # type: Optional[str]
        elif "serviceAccountName" in pod_spec and (
            run_config.job_template or run_config.job_template_path
        ):
            # On run-config job-template, no override
            service_account_name = None
        else:
            # Use agent value, if provided
            service_account_name = self.service_account_name
        if service_account_name is not None:
            pod_spec["serviceAccountName"] = service_account_name

        # Configure `image_pull_secrets` if specified
        if run_config.image_pull_secrets is not None:
            # On run-config, always override
            image_pull_secrets = (
                run_config.image_pull_secrets
            )  # type: Optional[Iterable[str]]
        elif "imagePullSecrets" in pod_spec and (
            run_config.job_template or run_config.job_template_path
        ):
            # On run-config job template, no override
            image_pull_secrets = None
        else:
            # Use agent, if provided
            image_pull_secrets = self.image_pull_secrets
        if image_pull_secrets is not None:
            pod_spec["imagePullSecrets"] = [{"name": s} for s in image_pull_secrets]

        # Default restartPolicy to Never
        _get_or_create(job, "spec.template.spec.restartPolicy", "Never")

        # Get the first container, which is used for the prefect job
        containers = _get_or_create(job, "spec.template.spec.containers", [])
        if not containers:
            containers.append({})
        container = containers[0]

        # Set container image
        container["image"] = image = get_flow_image(
            flow_run, default=container.get("image")
        )

        # set the the kubernetes imagePullPolicy, if image_pull_policy was specified
        if run_config.image_pull_policy is not None:
            container["imagePullPolicy"] = run_config.image_pull_policy

        # Set flow run command
        container["args"] = get_flow_run_command(flow_run).split()

        # Populate environment variables from the following sources,
        # with precedence:
        # - Values required for flow execution, hardcoded below
        # - Values set on the KubernetesRun object
        # - Values set using the `--env` CLI flag on the agent
        # - Values in the job template
        env = {"PREFECT__LOGGING__LEVEL": config.logging.level}
        env.update(self.env_vars)
        if run_config.env:
            env.update(run_config.env)
        env.update(
            {
                "PREFECT__BACKEND": config.backend,
                "PREFECT__CLOUD__AGENT__LABELS": str(self.labels),
                "PREFECT__CLOUD__API": config.cloud.api,
                "PREFECT__CLOUD__AUTH_TOKEN": config.cloud.agent.auth_token,
                "PREFECT__CLOUD__USE_LOCAL_SECRETS": "false",
                "PREFECT__CONTEXT__FLOW_RUN_ID": flow_run.id,
                "PREFECT__CONTEXT__FLOW_ID": flow_run.flow.id,
                "PREFECT__CONTEXT__IMAGE": image,
                "PREFECT__LOGGING__LOG_TO_CLOUD": str(self.log_to_cloud).lower(),
                "PREFECT__ENGINE__FLOW_RUNNER__DEFAULT_CLASS": "prefect.engine.cloud.CloudFlowRunner",
                "PREFECT__ENGINE__TASK_RUNNER__DEFAULT_CLASS": "prefect.engine.cloud.CloudTaskRunner",
            }
        )
        container_env = [{"name": k, "value": v} for k, v in env.items()]
        for entry in container.get("env", []):
            if entry["name"] not in env:
                container_env.append(entry)
        container["env"] = container_env

        # Set resource requirements if provided
        _get_or_create(container, "resources.requests")
        _get_or_create(container, "resources.limits")
        resources = container["resources"]
        if run_config.memory_request:
            resources["requests"]["memory"] = run_config.memory_request
        if run_config.memory_limit:
            resources["limits"]["memory"] = run_config.memory_limit
        if run_config.cpu_request:
            resources["requests"]["cpu"] = run_config.cpu_request
        if run_config.cpu_limit:
            resources["limits"]["cpu"] = run_config.cpu_limit

        return job

    @staticmethod
    def generate_deployment_yaml(
        token: str = None,
        api: str = None,
        namespace: str = None,
        image_pull_secrets: str = None,
        rbac: bool = False,
        latest: bool = False,
        mem_request: str = None,
        mem_limit: str = None,
        cpu_request: str = None,
        cpu_limit: str = None,
        image_pull_policy: str = None,
        service_account_name: str = None,
        labels: Iterable[str] = None,
        env_vars: dict = None,
        backend: str = None,
    ) -> str:
        """
        Generate and output an installable YAML spec for the agent.

        Args:
            - token (str, optional): A `RUNNER` token to give the agent
            - api (str, optional): A URL pointing to the Prefect API. Defaults to
                `https://api.prefect.io`
            - namespace (str, optional): The namespace to create Prefect jobs in. Defaults
                to `default`
            - image_pull_secrets (str, optional): The name of an image pull secret to use
                for Prefect jobs
            - rbac (bool, optional): Whether to include default RBAC configuration as
                part of the YAML. Defaults to `False`
            - latest (bool, optional): Whether to use the `latest` Prefect image.
                Defaults to `False`
            - mem_request (str, optional): Requested memory for Prefect init job.
            - mem_limit (str, optional): Limit memory for Prefect init job.
            - cpu_request (str, optional): Requested CPU for Prefect init job.
            - cpu_limit (str, optional): Limit CPU for Prefect init job.
            - image_pull_policy (str, optional): imagePullPolicy to use for Prefect init job.
                Job defaults to `IfNotPresent`.
            - service_account_name (str, optional): Name of a service account to use for
                Prefect init job. Job defaults to using `default` service account.
            - labels (List[str], optional): a list of labels, which are arbitrary string
                identifiers used by Prefect Agents when polling for work
            - env_vars (dict, optional): additional environment variables to attach to all
                jobs created by this agent and to set in the agent's own environment
            - backend (str, optional): toggle which backend to use for this agent.
                Defaults to backend currently set in config.

        Returns:
            - str: A string representation of the generated YAML
        """

        # Use defaults if not provided
        token = token or ""
        api = api or "https://api.prefect.io"
        namespace = namespace or "default"
        labels = labels or []
        mem_request = mem_request or ""
        mem_limit = mem_limit or ""
        cpu_request = cpu_request or ""
        cpu_limit = cpu_limit or ""
        image_pull_policy = image_pull_policy or ""
        service_account_name = service_account_name or ""
        backend = backend or config.backend

        version = prefect.__version__.split("+")
        image_version = (
            "latest" if len(version) > 1 or latest else (version[0] + "-python3.6")
        )

        with open(
            os.path.join(os.path.dirname(__file__), "deployment.yaml"), "r"
        ) as deployment_file:
            deployment = yaml.safe_load(deployment_file)

        agent_env = deployment["spec"]["template"]["spec"]["containers"][0]["env"]

        # Populate env vars
        agent_env[0]["value"] = token
        agent_env[1]["value"] = api
        agent_env[2]["value"] = namespace
        agent_env[3]["value"] = image_pull_secrets or ""
        agent_env[4]["value"] = str(labels)
        agent_env[11]["value"] = backend

        # Populate job resource env vars
        agent_env[5]["value"] = mem_request
        agent_env[6]["value"] = mem_limit
        agent_env[7]["value"] = cpu_request
        agent_env[8]["value"] = cpu_limit
        agent_env[9]["value"] = image_pull_policy
        agent_env[10]["value"] = service_account_name

        if env_vars:
            agent_env.append(
                {
                    "name": "PREFECT__CLOUD__AGENT__ENV_VARS",
                    "value": json.dumps(env_vars),
                }
            )

        # Use local prefect version for image
        deployment["spec"]["template"]["spec"]["containers"][0][
            "image"
        ] = "prefecthq/prefect:{}".format(image_version)

        # Populate image pull secrets if provided
        if image_pull_secrets:
            agent_env = deployment["spec"]["template"]["spec"]["imagePullSecrets"][0][
                "name"
            ] = image_pull_secrets
        else:
            del deployment["spec"]["template"]["spec"]["imagePullSecrets"]

        # Load RBAC if specified
        rbac_yaml = []
        if rbac:
            with open(
                os.path.join(os.path.dirname(__file__), "rbac.yaml"), "r"
            ) as rbac_file:
                rbac_generator = yaml.safe_load_all(rbac_file)

                for document in rbac_generator:
                    document["metadata"]["namespace"] = namespace
                    rbac_yaml.append(document)

        output_yaml = [deployment]
        output_yaml.extend(rbac_yaml)
        return yaml.safe_dump_all(output_yaml, explicit_start=True)


if __name__ == "__main__":
    KubernetesAgent().start()



import warnings
from typing import Any

from prefect.storage import Storage as _Storage


class _DeprecatedStorageMixin:
    def __new__(cls, *args: Any, **kwargs: Any) -> "_DeprecatedStorageMixin":
        if cls.__module__.startswith("prefect."):
            name = cls.__name__
            warnings.warn(
                f"`prefect.environments.storage.{name}` is deprecated, please use "
                f"`prefect.storage.{name}` instead",
                stacklevel=2,
            )
        return super().__new__(cls)


class Storage(_Storage, _DeprecatedStorageMixin):
    pass



from typing import Iterable

from prefect.run_configs.base import RunConfig


class DockerRun(RunConfig):
    """Configure a flow-run to run as a Docker container.

    Args:
        - image (str, optional): The image to use
        - env (dict, optional): Additional environment variables to set in the
            container
        - labels (Iterable[str], optional): an iterable of labels to apply to this
            run config. Labels are string identifiers used by Prefect Agents
            for selecting valid flow runs when polling for work

    Examples:

    Use the defaults set on the agent:

    ```python
    flow.run_config = DockerRun()
    ```

    Set an environment variable in the flow run process:

    ```python
    flow.run_config = DockerRun(env={"SOME_VAR": "value"})
    ```
    """

    def __init__(
        self, *, image: str = None, env: dict = None, labels: Iterable[str] = None
    ) -> None:
        super().__init__(env=env, labels=labels)
        self.image = image



import ssl
import smtplib
from email.message import EmailMessage
from typing import Any, cast

from prefect import Task
from prefect.client import Secret
from prefect.utilities.tasks import defaults_from_attrs


class EmailTask(Task):
    """
    Task for sending email from an authenticated email service over SMTP. For this task to
    function properly, you must have the `"EMAIL_USERNAME"` and `"EMAIL_PASSWORD"` Prefect
    Secrets set.  It is recommended you use a [Google App
    Password](https://support.google.com/accounts/answer/185833) if you use Gmail.  The default
    SMTP server is set to the Gmail SMTP server on port 465 (SMTP-over-SSL). Sending messages
    containing HTML code is supported - the default MIME type is set to the text/html.

    Args:
        - subject (str, optional): the subject of the email; can also be provided at runtime
        - msg (str, optional): the contents of the email, added as html; can be used in
            combination of msg_plain; can also be provided at runtime
        - email_to (str, optional): the destination email address to send the message to; can also
            be provided at runtime
        - email_from (str, optional): the email address to send from; defaults to
            notifications@prefect.io
        - smtp_server (str, optional): the hostname of the SMTP server; defaults to smtp.gmail.com
        - smtp_port (int, optional): the port number of the SMTP server; defaults to 465
        - smtp_type (str, optional): either SSL or STARTTLS; defaults to SSL
        - msg_plain (str, optional): the contents of the email, added as plain text can be used in
            combination of msg; can also be provided at runtime
        - email_to_cc (str, optional): additional email address to send the message to as cc;
            can also be provided at runtime
        - email_to_bcc (str, optional): additional email address to send the message to as bcc;
            can also be provided at runtime
        - **kwargs (Any, optional): additional keyword arguments to pass to the base Task
            initialization
    """

    def __init__(
        self,
        subject: str = None,
        msg: str = None,
        email_to: str = None,
        email_from: str = "notifications@prefect.io",
        smtp_server: str = "smtp.gmail.com",
        smtp_port: int = 465,
        smtp_type: str = "SSL",
        msg_plain: str = None,
        email_to_cc: str = None,
        email_to_bcc: str = None,
        **kwargs: Any,
    ):
        self.subject = subject
        self.msg = msg
        self.email_to = email_to
        self.email_from = email_from
        self.smtp_server = smtp_server
        self.smtp_port = smtp_port
        self.smtp_type = smtp_type
        self.msg_plain = msg_plain
        self.email_to_cc = email_to_cc
        self.email_to_bcc = email_to_bcc
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "subject",
        "msg",
        "email_to",
        "email_from",
        "smtp_server",
        "smtp_port",
        "smtp_type",
        "msg_plain",
        "email_to_cc",
        "email_to_bcc",
    )
    def run(
        self,
        subject: str = None,
        msg: str = None,
        email_to: str = None,
        email_from: str = None,
        smtp_server: str = None,
        smtp_port: int = None,
        smtp_type: str = None,
        msg_plain: str = None,
        email_to_cc: str = None,
        email_to_bcc: str = None,
    ) -> None:
        """
        Run method which sends an email.

        Args:
            - subject (str, optional): the subject of the email; defaults to the one provided
                at initialization
            - msg (str, optional): the contents of the email; defaults to the one provided
                at initialization
            - email_to (str, optional): the destination email address to send the message to;
                defaults to the one provided at initialization
            - email_from (str, optional): the email address to send from; defaults to the one
                provided at initialization
            - smtp_server (str, optional): the hostname of the SMTP server; defaults to the one
                provided at initialization
            - smtp_port (int, optional): the port number of the SMTP server; defaults to the one
                provided at initialization
            - smtp_type (str, optional): either SSL or STARTTLS; defaults to the one provided
                at initialization
            - msg_plain (str, optional): the contents of the email, added as plain text can be used in
                combination of msg; defaults to the one provided at initialization
            - email_to_cc (str, optional): additional email address to send the message to as cc;
                defaults to the one provided at initialization
            - email_to_bcc (str, optional): additional email address to send the message to as bcc;
                defaults to the one provided at initialization

        Returns:
            - None
        """

        username = cast(str, Secret("EMAIL_USERNAME").get())
        password = cast(str, Secret("EMAIL_PASSWORD").get())

        message = EmailMessage()
        message["Subject"] = subject
        message["From"] = email_from
        message["To"] = email_to
        if email_to_cc:
            message["Cc"] = email_to_cc
        if email_to_bcc:
            message["Bcc"] = email_to_bcc

        # https://docs.python.org/3/library/email.examples.html
        # If present, first set the plain content of the email. After which the html version is
        # added. This converts the message into a multipart/alternative container, with the
        # original text message as the first part and the new html message as the second part.
        if msg_plain:
            message.set_content(msg_plain, subtype="plain")
        if msg:
            message.add_alternative(msg, subtype="html")

        context = ssl.create_default_context()
        if smtp_type == "SSL":
            server = smtplib.SMTP_SSL(smtp_server, smtp_port, context=context)
        elif smtp_type == "STARTTLS":
            server = smtplib.SMTP(smtp_server, smtp_port)
            server.starttls(context=context)
        else:
            raise ValueError(f"{smtp_type} is an unsupported value for smtp_type.")

        server.login(username, password)
        try:
            server.send_message(message)
        finally:
            server.quit()



"""
This module contains a collection of tasks to run Data Quality tests using soda-sql library
"""

try:
    from prefect.tasks.sodasql.sodasql_tasks import SodaSQLScan
except ImportError as err:
    raise ImportError(
        'Using `prefect.tasks.sodasql` requires Prefect to be installed with the "sodasql" extra.'
    ) from err



from prefect.utilities.notifications.notifications import callback_factory
from prefect.utilities.notifications.notifications import slack_notifier
from prefect.utilities.notifications.notifications import gmail_notifier
from prefect.utilities.notifications.notifications import slack_message_formatter
from prefect.utilities.notifications.jira_notification import jira_notifier



import binascii
import importlib
import json
import sys
import warnings
from operator import attrgetter
from typing import TYPE_CHECKING
from distutils.version import LooseVersion

import cloudpickle

import prefect
from prefect.utilities.exceptions import StorageError

if TYPE_CHECKING:
    from prefect.core.flow import Flow  # pylint: disable=W0611


def get_flow_image(flow: "Flow") -> str:
    """
    Retrieve the image to use for this flow deployment. Will start by looking for
    an `image` value in the flow's `environment.metadata`. If not found then it will fall
    back to using the `flow.storage`.

    Args:
        - flow (Flow): A flow object

    Returns:
        - str: a full image name to use for this flow run

    Raises:
        - ValueError: if deployment attempted on unsupported Storage type and `image` not
            present in environment metadata
    """
    environment = flow.environment
    if (
        environment is not None
        and hasattr(environment, "metadata")
        and environment.metadata.get("image")
    ):
        return environment.metadata.get("image", "")
    else:
        storage = flow.storage
        if not isinstance(storage, prefect.storage.Docker):
            raise ValueError(
                f"Storage for flow run {flow.name} is not of type Docker and "
                f"environment has no `image` attribute in the metadata field."
            )

        return storage.name


def extract_flow_from_file(
    file_path: str = None, file_contents: str = None, flow_name: str = None
) -> "Flow":
    """
    Extract a flow object from a file.

    Args:
        - file_path (str, optional): A file path pointing to a .py file containing a flow
        - file_contents (str, optional): The string contents of a .py file containing a flow
        - flow_name (str, optional): A specific name of a flow to extract from a file.
            If not set then the first flow object retrieved from file will be returned.

    Returns:
        - Flow: A flow object extracted from a file

    Raises:
        - ValueError: if both `file_path` and `file_contents` are provided or neither are.
    """
    if file_path is not None:
        if file_contents is not None:
            raise ValueError(
                "Provide either `file_path` or `file_contents` but not both."
            )

        with open(file_path, "r") as f:
            contents = f.read()
    elif file_contents is not None:
        contents = file_contents
    else:
        raise ValueError("Provide either `file_path` or `file_contents`.")

    # Load objects from file into dict
    exec_vals = {}  # type: ignore
    exec(contents, exec_vals)

    # Grab flow name from values loaded via exec
    flows = {o.name: o for o in exec_vals.values() if isinstance(o, prefect.Flow)}
    if flows:
        if flow_name:
            if flow_name in flows:
                return flows[flow_name]
            else:
                flows_list = "\n".join("- %r" % n for n in sorted(flows))
                raise ValueError(
                    f"Flow {flow_name!r} not found in file. Found flows:\n{flows_list}"
                )
        else:
            return list(flows.values())[0]
    else:
        raise ValueError("No flows found in file.")


def extract_flow_from_module(module_str: str, flow_name: str = None) -> "Flow":
    """
    Extract a flow object from a python module.

    Args:
        - module_str (str): A module path pointing to a .py file containing a flow.
            For example, 'myrepo.mymodule.myflow' where myflow.py contains the flow.
            Additionally, `:` can be used to access module's attribute, for example,
            'myrepo.mymodule.myflow:flow' or 'myrepo.mymodule.myflow:MyObj.newflow'.
        - flow_name (str, optional): A specific name of a flow to extract from a file.
            If not provided, the `module_str` must have an attribute specifier
            or only one `Flow` object must be present in the module, otherwise
            an error will be raised.

    Returns:
        - Flow: A flow object extracted from a file
    """
    mod_name, obj_name_present, obj_name = module_str.partition(":")

    module = importlib.import_module(mod_name)

    if obj_name_present:
        try:
            flow = attrgetter(obj_name)(module)
        except AttributeError:
            raise ValueError(f"Failed to find flow at {module_str!r}") from None
        if callable(flow):
            flow = flow()
        if not isinstance(flow, prefect.Flow):
            raise TypeError(
                f"Object at {module_str!r} is a {type(flow)} not a `prefect.Flow`"
            )
        if flow_name is not None and flow.name != flow_name:
            raise ValueError(
                f"Flow at {module_str!r} is named {flow.name!r}, expected {flow_name!r}"
            )
        return flow
    else:
        flows = {}
        for attr in dir(module):
            obj = getattr(module, attr, None)
            if isinstance(obj, prefect.Flow):
                flows[obj.name] = obj
        if flow_name is not None:
            if flow_name in flows:
                return flows[flow_name]
            raise ValueError(f"Failed to find flow {flow_name!r} in {module_str!r}")
        elif len(flows) == 1:
            return flows.popitem()[1]
        elif len(flows) > 1:
            raise ValueError(
                "Multiple flows found in {module_str!r}, please provide `flow_name` to select one."
            )
        else:
            raise ValueError("No flows found in {module_str!r}")


def flow_to_bytes_pickle(flow: "Flow") -> bytes:
    """Serialize a flow to bytes.

    The flow is serialized using `cloudpickle`, with some extra metadata on
    included via JSON. The flow can be reloaded using `flow_from_bytes_pickle`.

    Args:
        - flow (Flow): the flow to be serialized.

    Returns:
        - bytes: a serialized representation of the flow.
    """
    flow_data = binascii.b2a_base64(
        cloudpickle.dumps(flow, protocol=4), newline=False
    ).decode("utf-8")
    out = json.dumps({"flow": flow_data, "versions": _get_versions()})
    return out.encode("utf-8")


def _get_versions() -> dict:
    """Get version info on libraries where a version-mismatch between
    registration and execution environment may cause a flow to fail to load
    properly"""
    return {
        "cloudpickle": cloudpickle.__version__,
        "prefect": prefect.__version__,
        "python": "%d.%d.%d" % sys.version_info[:3],
    }


def flow_from_bytes_pickle(data: bytes) -> "Flow":
    """Load a flow from bytes."""
    try:
        info = json.loads(data.decode("utf-8"))
    except Exception:
        # Serialized using older version of prefect, use cloudpickle directly
        flow_bytes = data
        reg_versions = {}
    else:
        flow_bytes = binascii.a2b_base64(info["flow"])
        reg_versions = info["versions"]

    run_versions = _get_versions()

    try:
        flow = cloudpickle.loads(flow_bytes)
    except Exception as exc:
        parts = ["An error occurred while unpickling the flow:", f"  {exc!r}"]
        # Check for mismatched versions to provide a better error if possible
        mismatches = []
        for name, v1 in sorted(reg_versions.items()):
            if name in run_versions:
                v2 = run_versions[name]
                if LooseVersion(v1) != v2:
                    mismatches.append(
                        f"  - {name}: (flow built with {v1!r}, currently running with {v2!r})"
                    )
        if mismatches:
            parts.append(
                "This may be due to one of the following version mismatches between "
                "the flow build and execution environments:"
            )
            parts.extend(mismatches)
        if isinstance(exc, ImportError):
            # If it's an import error, also note that the user may need to package
            # their dependencies
            prefix = "This also may" if mismatches else "This may"
            parts.append(
                f"{prefix} be due to a missing Python module in your current "
                "environment. Please ensure you have all required flow "
                "dependencies installed."
            )
        raise StorageError("\n".join(parts)) from exc

    run_prefect = run_versions["prefect"]
    reg_prefect = reg_versions.get("prefect")
    if reg_prefect and LooseVersion(reg_prefect) != run_prefect:
        # If we didn't error above, still check that the prefect versions match
        # and warn if they don't. Prefect version mismatches *may* work, but
        # they may also error later leading to confusing behavior.
        warnings.warn(
            f"This flow was built using Prefect {reg_prefect!r}, but you currently "
            f"have Prefect {run_prefect!r} installed. We recommend loading flows "
            "with the same Prefect version they were built with, failure to do so "
            "may result in errors."
        )
    return flow



import requests
import logging

from time import sleep
from urllib.parse import urlparse

from requests import exceptions as requests_exceptions
from requests.auth import AuthBase

import prefect

from prefect.utilities.exceptions import PrefectError

RESTART_CLUSTER_ENDPOINT = ("POST", "api/2.0/clusters/restart")

START_CLUSTER_ENDPOINT = ("POST", "api/2.0/clusters/start")

TERMINATE_CLUSTER_ENDPOINT = ("POST", "api/2.0/clusters/delete")


RUN_NOW_ENDPOINT = ("POST", "api/2.0/jobs/run-now")

SUBMIT_RUN_ENDPOINT = ("POST", "api/2.0/jobs/runs/submit")

GET_RUN_ENDPOINT = ("GET", "api/2.0/jobs/runs/get")

CANCEL_RUN_ENDPOINT = ("POST", "api/2.0/jobs/runs/cancel")

USER_AGENT_HEADER = {"user-agent": "prefect-{v}".format(v=prefect.__version__)}


class RunState:
    """
    Utility class for the run state concept of Databricks runs.
    """

    def __init__(self, life_cycle_state, result_state, state_message):
        self.life_cycle_state = life_cycle_state
        self.result_state = result_state
        self.state_message = state_message

    @property
    def is_terminal(self) -> bool:
        """True if the current state is a terminal state."""
        if self.life_cycle_state not in RUN_LIFE_CYCLE_STATES:
            raise PrefectError(
                (
                    "Unexpected life cycle state: {}: If the state has "
                    "been introduced recently, please check the Databricks user "
                    "guide for troubleshooting information"
                ).format(self.life_cycle_state)
            )
        return self.life_cycle_state in ("TERMINATED", "SKIPPED", "INTERNAL_ERROR")

    @property
    def is_successful(self) -> bool:
        """True if the result state is SUCCESS"""
        return self.result_state == "SUCCESS"

    def __eq__(self, other):
        return (
            self.life_cycle_state == other.life_cycle_state
            and self.result_state == other.result_state
            and self.state_message == other.state_message
        )

    def __repr__(self):
        return str(self.__dict__)


# noinspection PyAbstractClass
class DatabricksHook:
    """
    Interact with Databricks.

    Args:
        - databricks_conn_id (dict): The name of the databricks connection to use.

        - timeout_seconds (int): The amount of time in seconds the requests library
            will wait before timing-out.

        - retry_limit (int): The number of times to retry the connection in case of
            service outages.

        - retry_delay (float): The number of seconds to wait between retries (it
            might be a floating point number).
    """

    def __init__(
        self,
        databricks_conn_id: dict,
        timeout_seconds=180,
        retry_limit=3,
        retry_delay=1.0,
    ):

        self.databricks_conn = databricks_conn_id
        self.timeout_seconds = timeout_seconds
        if retry_limit < 1:
            raise ValueError("Retry limit must be greater than equal to 1")
        self.retry_limit = retry_limit
        self.retry_delay = retry_delay

    @staticmethod
    def _parse_host(host):
        """
        The purpose of this function is to be robust to improper connections
        settings provided by users, specifically in the host field.

        For example -- when users supply `https://xx.cloud.databricks.com` as the
        host, we must strip out the protocol to get the host.::

            h = DatabricksHook()
            assert h._parse_host('https://xx.cloud.databricks.com') == \
                'xx.cloud.databricks.com'

        In the case where users supply the correct ``xx.cloud.databricks.com`` as the
        host, this function is a no-op.::

            assert h._parse_host('xx.cloud.databricks.com') == 'xx.cloud.databricks.com'

        """
        urlparse_host = urlparse(host).hostname
        if urlparse_host:
            # In this case, host = https://xx.cloud.databricks.com
            return urlparse_host
        else:
            # In this case, host = xx.cloud.databricks.com
            return host

    def _do_api_call(self, endpoint_info, json):
        """
        Utility function to perform an API call with retries

        Args:
            - endpoint_info (tuple[string, string]): Tuple of method and endpoint

            - json (dict): Parameters for this API call.

        Returns:

            dict: If the api call returns a OK status code,
                this function returns the response in JSON. Otherwise,
                we throw an PrefectError.
        """
        method, endpoint = endpoint_info

        if "token" in self.databricks_conn:
            logging.info("Using token auth. ")
            auth = _TokenAuth(self.databricks_conn["token"])
            host = self._parse_host(self.databricks_conn["host"])
        else:
            logging.info("Using basic auth. ")
            auth = (self.databricks_conn["login"], self.databricks_conn["password"])
            host = self.databricks_conn["host"]

        url = "https://{host}/{endpoint}".format(
            host=self._parse_host(host), endpoint=endpoint
        )

        if method == "GET":
            request_func = requests.get
        elif method == "POST":
            request_func = requests.post
        elif method == "PATCH":
            request_func = requests.patch
        else:
            raise PrefectError("Unexpected HTTP Method: " + method)

        attempt_num = 1
        while True:
            try:
                response = request_func(
                    url,
                    json=json,
                    auth=auth,
                    headers=USER_AGENT_HEADER,
                    timeout=self.timeout_seconds,
                )
                response.raise_for_status()
                return response.json()
            except requests_exceptions.RequestException as e:
                if not _retryable_error(e):
                    # In this case, the user probably made a mistake.
                    # Don't retry.
                    raise PrefectError(
                        "Response: {0}, Status Code: {1}".format(
                            e.response.content, e.response.status_code
                        )
                    ) from e

                self._log_request_error(attempt_num, e)

            if attempt_num == self.retry_limit:
                raise PrefectError(
                    (
                        "API requests to Databricks failed {} times. " + "Giving up."
                    ).format(self.retry_limit)
                )

            attempt_num += 1
            sleep(self.retry_delay)

    def _log_request_error(self, attempt_num, error):
        logging.error(
            "Attempt %s API Request to Databricks failed with reason: %s",
            attempt_num,
            error,
        )

    def run_now(self, json):
        """
        Utility function to call the ``api/2.0/jobs/run-now`` endpoint.

        Args:
         - json (dict): The data used in the body of the request to the ``run-now`` endpoint.

        Returns:
            - str: the run_id as a string
        """
        response = self._do_api_call(RUN_NOW_ENDPOINT, json)
        return response["run_id"]

    def submit_run(self, json):
        """
        Utility function to call the ``api/2.0/jobs/runs/submit`` endpoint.

        Args:
            - json (dict): The data used in the body of the request to the ``submit`` endpoint.

        Returns:
            - str: the run_id as a string
        """
        response = self._do_api_call(SUBMIT_RUN_ENDPOINT, json)
        return response["run_id"]

    def get_run_page_url(self, run_id: str) -> str:
        """
        Retrieves run_page_url.

        Args:
            - run_id (str): id of the run

        Returns
            - str: URL of the run page
        """
        json = {"run_id": run_id}
        response = self._do_api_call(GET_RUN_ENDPOINT, json)
        return response["run_page_url"]

    def get_job_id(self, run_id: str) -> str:
        """
        Retrieves job_id from run_id.

        Args:
            - run_id (str): id of the run

        Returns
            - str: Job id for given Databricks run
        """
        json = {"run_id": run_id}
        response = self._do_api_call(GET_RUN_ENDPOINT, json)
        return response["job_id"]

    def get_run_state(self, run_id: str) -> RunState:
        """
        Retrieves run state of the run.

        Args:
            - run_id (str): id of the run

        Returns
            - str: state of the run
        """
        json = {"run_id": run_id}
        response = self._do_api_call(GET_RUN_ENDPOINT, json)
        state = response["state"]
        life_cycle_state = state["life_cycle_state"]
        # result_state may not be in the state if not terminal
        result_state = state.get("result_state", None)
        state_message = state["state_message"]
        return RunState(life_cycle_state, result_state, state_message)

    def cancel_run(self, run_id: str) -> None:
        """
        Cancels the run.

        Args:
            - run_id (str): id of the run
        """
        json = {"run_id": run_id}
        self._do_api_call(CANCEL_RUN_ENDPOINT, json)

    def restart_cluster(self, json: dict) -> None:
        """
        Restarts the cluster.

        Args:
            - json (dict): json dictionary containing cluster specification.
        """
        self._do_api_call(RESTART_CLUSTER_ENDPOINT, json)

    def start_cluster(self, json: dict) -> None:
        """
        Starts the cluster.

        Args:
            - json (dict): json dictionary containing cluster specification.
        """
        self._do_api_call(START_CLUSTER_ENDPOINT, json)

    def terminate_cluster(self, json: dict) -> None:
        """
        Terminates the cluster.

        Args:
            - json (dict): json dictionary containing cluster specification.
        """
        self._do_api_call(TERMINATE_CLUSTER_ENDPOINT, json)


def _retryable_error(exception):
    return (
        isinstance(
            exception,
            (requests_exceptions.ConnectionError, requests_exceptions.Timeout),
        )
        or exception.response is not None
        and exception.response.status_code >= 500
    )


RUN_LIFE_CYCLE_STATES = [
    "PENDING",
    "RUNNING",
    "TERMINATING",
    "TERMINATED",
    "SKIPPED",
    "INTERNAL_ERROR",
]


class _TokenAuth(AuthBase):
    """
    Helper class for requests Auth field. AuthBase requires you to implement the __call__
    magic function.
    """

    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        r.headers["Authorization"] = "Bearer " + self.token
        return r



from typing import Any

from marshmallow import fields, post_load

from prefect.storage import (
    Azure,
    Bitbucket,
    CodeCommit,
    Docker,
    GCS,
    GitHub,
    GitLab,
    Local,
    Module,
    S3,
    Storage,
    Webhook,
    Git,
)
from prefect.utilities.serialization import JSONCompatible, ObjectSchema, OneOfSchema


class BaseStorageSchema(ObjectSchema):
    class Meta:
        object_class = Storage

    flows = fields.Dict(key=fields.Str(), values=fields.Str())
    secrets = fields.List(fields.Str(), allow_none=True)

    @post_load
    def create_object(self, data: dict, **kwargs: Any) -> Storage:
        flows = data.pop("flows", {})
        base_obj = super().create_object(data, **kwargs)
        base_obj.flows = flows
        return base_obj


class AzureSchema(BaseStorageSchema):
    class Meta:
        object_class = Azure

    container = fields.String(allow_none=False)
    blob_name = fields.String(allow_none=True)
    stored_as_script = fields.Bool(allow_none=True)


class DockerSchema(BaseStorageSchema):
    class Meta:
        object_class = Docker

    registry_url = fields.String(allow_none=True)
    image_name = fields.String(allow_none=True)
    image_tag = fields.String(allow_none=True)
    path = fields.Str(allow_none=True)
    stored_as_script = fields.Bool(allow_none=True)
    prefect_version = fields.String(allow_none=False)


class GCSSchema(BaseStorageSchema):
    class Meta:
        object_class = GCS

    bucket = fields.Str(allow_none=False)
    key = fields.Str(allow_none=True)
    project = fields.Str(allow_none=True)
    stored_as_script = fields.Bool(allow_none=True)


class LocalSchema(BaseStorageSchema):
    class Meta:
        object_class = Local

    directory = fields.Str(allow_none=False)
    path = fields.Str(allow_none=True)
    stored_as_script = fields.Bool(allow_none=True)

    @post_load
    def create_object(self, data: dict, **kwargs: Any) -> Storage:
        return super().create_object({"validate": False, **data}, **kwargs)


class S3Schema(BaseStorageSchema):
    class Meta:
        object_class = S3

    bucket = fields.String(allow_none=False)
    key = fields.String(allow_none=True)
    stored_as_script = fields.Bool(allow_none=True)
    client_options = fields.Dict(
        key=fields.Str(), values=JSONCompatible(), allow_none=True
    )


class GitHubSchema(BaseStorageSchema):
    class Meta:
        object_class = GitHub

    repo = fields.String(allow_none=False)
    ref = fields.String(allow_none=True)
    path = fields.String(allow_none=False)
    access_token_secret = fields.String(allow_none=True)
    base_url = fields.String(allow_none=True)


class GitLabSchema(BaseStorageSchema):
    class Meta:
        object_class = GitLab

    repo = fields.String(allow_none=False)
    path = fields.String(allow_none=True)
    host = fields.String(allow_none=True)
    ref = fields.String(allow_none=True)
    access_token_secret = fields.String(allow_none=True)


class BitbucketSchema(BaseStorageSchema):
    class Meta:
        object_class = Bitbucket

    project = fields.String(allow_none=False)
    repo = fields.String(allow_none=False)
    workspace = fields.String(allow_none=True)
    host = fields.String(allow_none=True)
    path = fields.String(allow_none=True)
    ref = fields.String(allow_none=True)
    access_token_secret = fields.String(allow_none=True)
    cloud_username_secret = fields.String(allow_none=True)
    cloud_app_password_secret = fields.String(allow_none=True)


class CodeCommitSchema(BaseStorageSchema):
    class Meta:
        object_class = CodeCommit

    repo = fields.String(allow_none=False)
    path = fields.String(allow_none=True)
    commit = fields.String(allow_none=True)
    client_options = fields.Dict(
        key=fields.Str(), values=JSONCompatible(), allow_none=True
    )


class WebhookSchema(BaseStorageSchema):
    class Meta:
        object_class = Webhook

    build_request_kwargs = fields.Dict(key=fields.Str, allow_none=False)
    build_request_http_method = fields.String(allow_none=False)
    get_flow_request_kwargs = fields.Dict(key=fields.Str, allow_none=False)
    get_flow_request_http_method = fields.String(allow_none=False)
    stored_as_script = fields.Bool(allow_none=True)


class GitSchema(BaseStorageSchema):
    class Meta:
        object_class = Git

    flow_path = fields.String(allow_none=False)
    repo = fields.String(allow_none=False)
    repo_host = fields.String(allow_none=False)
    flow_name = fields.String(allow_none=True)
    git_token_secret_name = fields.String(allow_none=True)
    git_token_username = fields.String(allow_none=True)
    branch_name = fields.String(allow_none=True)
    tag = fields.String(allow_none=True)
    commit = fields.String(allow_none=True)
    clone_depth = fields.Integer(allow_none=True)
    use_ssh = fields.Boolean(allow_none=False)
    format_access_token = fields.Boolean(allow_none=False)


class ModuleSchema(BaseStorageSchema):
    class Meta:
        object_class = Module

    module = fields.String(allow_none=False)


class StorageSchema(OneOfSchema):
    """
    Field that chooses between several nested schemas
    """

    # map class name to schema
    type_schemas = {
        "Azure": AzureSchema,
        "Bitbucket": BitbucketSchema,
        "CodeCommit": CodeCommitSchema,
        "Docker": DockerSchema,
        "GCS": GCSSchema,
        "GitHub": GitHubSchema,
        "GitLab": GitLabSchema,
        "Local": LocalSchema,
        "Module": ModuleSchema,
        "S3": S3Schema,
        "Storage": BaseStorageSchema,
        "Webhook": WebhookSchema,
        "Git": GitSchema,
    }



import copy
import cloudpickle
import itertools
import multiprocessing
import os
import signal
import subprocess
import sys
import threading
import warnings
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import TimeoutError as FutureTimeout
from functools import wraps
from logging import Logger
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Union, Sequence, Mapping

import prefect
from prefect.utilities.exceptions import TaskTimeoutError
from prefect.utilities.logging import get_logger

if TYPE_CHECKING:
    import prefect.engine.runner
    import prefect.engine.state
    from prefect.core.edge import Edge
    from prefect.core.task import Task
    from prefect.engine.state import State


StateList = Union["State", List["State"]]


def run_with_heartbeat(
    runner_method: Callable[..., "prefect.engine.state.State"]
) -> Callable[..., "prefect.engine.state.State"]:
    """
    Utility decorator for running class methods with a heartbeat.  The class should implement
    `self._heartbeat` with no arguments.
    """

    @wraps(runner_method)
    def inner(
        self: "prefect.engine.runner.Runner", *args: Any, **kwargs: Any
    ) -> "prefect.engine.state.State":
        try:
            p = None
            try:
                if self._heartbeat():
                    # we use Popen + a prefect CLI for a few reasons:
                    # - using threads would interfere with the task; for example, a task
                    #   which does not release the GIL would prevent the heartbeat thread from
                    #   firing
                    # - using multiprocessing.Process would release the GIL but a subprocess
                    #   cannot be spawned from a daemonic subprocess, and Dask sometimes will
                    #   submit tasks to run within daemonic subprocesses
                    current_env = dict(os.environ).copy()
                    auth_token = prefect.context.config.cloud.get("auth_token")
                    api_url = prefect.context.config.cloud.get("api")
                    current_env.setdefault("PREFECT__CLOUD__AUTH_TOKEN", auth_token)
                    current_env.setdefault("PREFECT__CLOUD__API", api_url)
                    clean_env = {k: v for k, v in current_env.items() if v is not None}
                    p = subprocess.Popen(
                        self.heartbeat_cmd,
                        env=clean_env,
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.DEVNULL,
                    )
            except Exception:
                self.logger.exception(
                    "Heartbeat failed to start.  This could result in a zombie run."
                )
            return runner_method(self, *args, **kwargs)
        finally:
            if p is not None:
                exit_code = p.poll()
                if exit_code is not None:
                    msg = "Heartbeat process died with exit code {}".format(exit_code)
                    self.logger.error(msg)
                p.kill()

    return inner


def run_with_thread_timeout(
    fn: Callable,
    args: Sequence = (),
    kwargs: Mapping = None,
    timeout: int = None,
    logger: Logger = None,
    name: str = None,
) -> Any:
    """
    Helper function for implementing timeouts on function executions.
    Implemented by setting a `signal` alarm on a timer. Must be run in the main thread.

    Args:
        - fn (callable): the function to execute
        - args (Sequence): arguments to pass to the function
        - kwargs (Mapping): keyword arguments to pass to the function
        - timeout (int): the length of time to allow for execution before raising a
            `TaskTimeoutError`, represented as an integer in seconds
        - logger (Logger): an optional logger to use. If not passed, a logger for the
            `prefect.executors.run_with_thread_timeout` namespace will be created.
        - name (str): an optional name to attach to logs for this function run, defaults
            to the name of the given function. Provides an interface for passing task
            names for logs.

    Returns:
        - the result of `fn(*args, **kwargs)`

    Raises:
        - TaskTimeoutError: if function execution exceeds the allowed timeout
        - ValueError: if run from outside the main thread
    """
    logger = logger or get_logger()
    name = name or f"Function '{fn.__name__}'"
    kwargs = kwargs or {}

    if timeout is None:
        return fn(*args, **kwargs)

    def error_handler(signum, frame):  # type: ignore
        raise TaskTimeoutError("Execution timed out.")

    try:
        # Set the signal handler for alarms
        signal.signal(signal.SIGALRM, error_handler)
        # Raise the alarm if `timeout` seconds pass
        logger.debug(f"{name}: Sending alarm with {timeout}s timeout...")
        signal.alarm(timeout)
        logger.debug(f"{name}: Executing function in main thread...")
        return fn(*args, **kwargs)
    finally:
        signal.alarm(0)


def multiprocessing_safe_run_and_retrieve(
    queue: multiprocessing.Queue,
    payload: bytes,
) -> None:
    """
    Gets the return value from a function and puts it in a multiprocessing-safe
    container. Helper function for `run_with_multiprocess_timeout`, must be defined
    top-level so it can be pickled and sent to `multiprocessing.Process`

    Passing the payload serialized allows us to escape the limitations of the python
    native pickler which will fail on tasks defined in scripts because of name
    mismatches. Whilst this particular example only affects the `func` arg, any of the
    others could be affected by other pickle limitations as well.

    Args:
        - queue (multiprocessing.Queue): The queue to pass the resulting payload to
        - payload (bytes): A serialized dictionary containing the data required to run
            the function. Should be serialized with `cloudpickle.dumps`
            Expects the following keys:
            - fn (Callable): The function to call
            - args (list): Positional argument values to call the function with
            - kwargs (Mapping): Keyword arguments to call the function with
            - context (dict): The prefect context dictionary to use during execution
            - name (str): an optional name to attach to logs for this function run,
                defaults to the name of the given function. Provides an interface for
                passing task names for logs.
            - logger (Logger): the logger to use
    """
    request = cloudpickle.loads(payload)

    fn: Callable = request["fn"]
    context: dict = request.get("context", {})
    args: Sequence = request.get("args", [])
    kwargs: Mapping = request.get("kwargs", {})
    name: str = request.get("name", f"Function '{fn.__name__}'")
    logger: Logger = request.get("logger") or get_logger()

    try:
        with prefect.context(context):
            logger.debug(f"{name}: Executing...")
            return_val = fn(*args, **kwargs)
            logger.debug(f"{name}: Execution successful.")
    except Exception as exc:
        return_val = exc
        logger.error(
            f"{name}: Encountered a {type(exc).__name__}, "
            f"returning details as a result..."
        )

    try:
        pickled_val = cloudpickle.dumps(return_val)
    except Exception as exc:
        err_msg = (
            f"Failed to pickle result of type {type(return_val).__name__!r} with "
            f'exception: "{type(exc).__name__}: {str(exc)}". This timeout handler "'
            "requires your function return value to be serializable with `cloudpickle`."
        )
        logger.error(f"{name}: {err_msg}")
        pickled_val = cloudpickle.dumps(RuntimeError(err_msg))

    logger.debug(f"{name}: Passing result back to main process...")

    try:
        queue.put(pickled_val)
    except Exception:
        logger.error(
            f"{name}: Failed to put result in queue to main process!",
            exc_info=True,
        )
        raise


def run_with_multiprocess_timeout(
    fn: Callable,
    args: Sequence = (),
    kwargs: Mapping = None,
    timeout: int = None,
    logger: Logger = None,
    name: str = None,
) -> Any:
    """
    Helper function for implementing timeouts on function executions.
    Implemented by spawning a new multiprocess.Process() and joining with timeout.

    Args:
        - fn (callable): the function to execute
        - args (Sequence): arguments to pass to the function
        - kwargs (Mapping): keyword arguments to pass to the function
        - timeout (int): the length of time to allow for execution before raising a
            `TaskTimeoutError`, represented as an integer in seconds
        - logger (Logger): an optional logger to use. If not passed, a logger for the
            `prefect.` namespace will be created.
        - name (str): an optional name to attach to logs for this function run, defaults
            to the name of the given function. Provides an interface for passing task
            names for logs.

    Returns:
        - the result of `f(*args, **kwargs)`

    Raises:
        - AssertionError: if run from a daemonic process
        - TaskTimeoutError: if function execution exceeds the allowed timeout
    """
    logger = logger or get_logger()
    name = name or f"Function '{fn.__name__}'"
    kwargs = kwargs or {}

    if timeout is None:
        return fn(*args, **kwargs)

    spawn_mp = multiprocessing.get_context("spawn")

    # Create a queue to pass the function return value back
    queue = spawn_mp.Queue()  # type: multiprocessing.Queue

    # Set internal kwargs for the helper function
    request = {
        "fn": fn,
        "args": args,
        "kwargs": kwargs,
        "context": prefect.context.to_dict(),
        "name": name,
        "logger": logger,
    }
    payload = cloudpickle.dumps(request)

    run_process = spawn_mp.Process(
        target=multiprocessing_safe_run_and_retrieve, args=(queue, payload)
    )
    logger.debug(f"{name}: Sending execution to a new process...")
    run_process.start()
    logger.debug(f"{name}: Waiting for process to return with {timeout}s timeout...")
    run_process.join(timeout)
    run_process.terminate()

    # Handle the process result, if the queue is empty the function did not finish
    # before the timeout
    logger.debug(f"{name}: Execution process closed, collecting result...")
    if not queue.empty():
        result = cloudpickle.loads(queue.get())
        if isinstance(result, Exception):
            raise result
        return result
    else:
        raise TaskTimeoutError(f"Execution timed out for {name}.")


def run_task_with_timeout(
    task: "Task",
    args: Sequence = (),
    kwargs: Mapping = None,
    logger: Logger = None,
) -> Any:
    """
    Helper function for implementing timeouts on task executions.

    The exact implementation varies depending on whether this function is being
    run in the main thread or a non-daemonic subprocess.  If this is run from a
    daemonic subprocess or on Windows, the task is run in a `ThreadPoolExecutor`
    and only a soft timeout is enforced, meaning a `TaskTimeoutError` is raised at the
    appropriate time but the task continues running in the background.

    The task is passed instead of a function so we can give better logs and messages.
    If you need to run generic functions with timeout handlers,
    `run_with_thread_timeout` or `run_with_multiprocess_timeout` can be called directly

    Args:
        - task (Task): the task to execute
            `task.timeout` specifies the number of seconds to allow `task.run` to run
            for before terminating
        - args (Sequence): arguments to pass to the function
        - kwargs (Mapping): keyword arguments to pass to the function
        - logger (Logger): an optional logger to use. If not passed, a logger for the
            `prefect.run_task_with_timeout_handler` namespace will be created.

    Returns:
        - the result of `f(*args, **kwargs)`

    Raises:
        - TaskTimeoutError: if function execution exceeds the allowed timeout
    """
    logger = logger or get_logger()
    name = prefect.context.get("task_full_name", task.name)
    kwargs = kwargs or {}

    # if no timeout, just run the function
    if task.timeout is None:
        return task.run(*args, **kwargs)  # type: ignore

    # if we are running the main thread, use a signal to stop execution at the
    # appropriate time; else if we are running in a non-daemonic process, spawn
    # a subprocess to kill at the appropriate time
    if not sys.platform.startswith("win"):

        if threading.current_thread() is threading.main_thread():
            # This case is typically encountered when using a non-parallel or local
            # multiprocess scheduler because then each worker is in the main
            # thread
            logger.debug(f"Task '{name}': Attaching thread based timeout handler...")
            return run_with_thread_timeout(
                task.run,
                args,
                kwargs,
                timeout=task.timeout,
                logger=logger,
                name=f"Task '{name}'",
            )

        elif multiprocessing.current_process().daemon is False:
            # This case is typically encountered when using a multithread distributed
            # executor
            logger.debug(f"Task '{name}': Attaching process based timeout handler...")
            return run_with_multiprocess_timeout(
                task.run,
                args,
                kwargs,
                timeout=task.timeout,
                logger=logger,
                name=f"Task '{name}'",
            )

        # We are in a daemonic process and cannot enforce a timeout
        # This case is typically encountered when using a multiprocess distributed
        # executor
        soft_timeout_reason = "in a daemonic subprocess"
    else:
        # We are in windows and cannot enforce a timeout
        soft_timeout_reason = "on Windows"

    msg = (
        f"This task is running {soft_timeout_reason}; "
        "consequently Prefect can only enforce a soft timeout limit, i.e., "
        "if your Task reaches its timeout limit it will enter a TimedOut state "
        "but continue running in the background."
    )

    logger.debug(
        f"Task '{name}': Falling back to daemonic soft limit timeout handler because "
        f"we are running {soft_timeout_reason}."
    )
    warnings.warn(msg, stacklevel=2)
    executor = ThreadPoolExecutor()

    def run_with_ctx(context: dict) -> Any:
        with prefect.context(context):
            return task.run(*args, **kwargs)  # type: ignore

    # Run the function in the background and then retrieve its result with a timeout
    fut = executor.submit(run_with_ctx, prefect.context.to_dict())

    try:
        return fut.result(timeout=task.timeout)
    except FutureTimeout as exc:
        raise TaskTimeoutError(
            f"Execution timed out but was executed {soft_timeout_reason} and will "
            "continue to run in the background."
        ) from exc


class RecursiveCall(Exception):
    def __init__(self, func: Callable, *args: Any, **kwargs: Any):
        self.func = func
        self.args = args
        self.kwargs = kwargs


def tail_recursive(func: Callable) -> Callable:
    """
    Helper function to facilitate tail recursion of the wrapped function.

    This allows for recursion with unlimited depth since a stack is not allocated for
    each "nested" call. Note: instead of calling the target function in question, a
    `RecursiveCall` exception must be raised instead.

    Args:
        - fn (callable): the function to execute

    Returns:
        - the result of `f(*args, **kwargs)`

    Raises:
        - RecursionError: if a recursive "call" (raised exception) is made with a function that is
            not decorated with `tail_recursive` decorator.
    """

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        while True:
            try:
                return func(*args, **kwargs)
            except RecursiveCall as exc:
                try:
                    call_func = getattr(exc.func, "__wrapped_func__")
                except AttributeError as attr_error:
                    raise RecursionError(
                        "function has not been wrapped to provide tail recursion (func={})".format(
                            exc.func
                        )
                    ) from attr_error

                # there may be multiple nested recursive calls, we should only
                # respond to calls for the wrapped function explicitly,
                # otherwise allow the call to continue to propagate
                if call_func != func:
                    raise exc
                args = exc.args
                kwargs = exc.kwargs
                continue

    setattr(wrapper, "__wrapped_func__", func)
    return wrapper


def prepare_upstream_states_for_mapping(
    state: "State",
    upstream_states: "Dict[Edge, State]",
    mapped_children: "Dict[Task, list]",
    executor: "prefect.executors.Executor",
) -> list:
    """
    If the task is being mapped, submits children tasks for execution. Returns a `Mapped` state.

    Args:
        - state (State): the parent task's current state
        - upstream_states (Dict[Edge, State]): the upstream states to this task
        - mapped_children (Dict[Task, List[State]]): any mapped children upstream of this task

    Returns:
        - List: a restructured list of upstream states correponding to each new mapped child task
    """

    # if the current state is failed / skipped or otherwise
    # in a state that signifies we should not continue with mapping,
    # we return an empty list
    if state.is_pending() or state.is_failed() or state.is_skipped():
        return []

    map_upstream_states = []

    # copy the mapped children dict to avoid mutating the global one
    # when we flatten mapped children
    mapped_children = mapped_children.copy()

    # we don't know how long the iterables are, but we want to iterate until we reach
    # the end of the shortest one
    counter = itertools.count()

    # preprocessing
    for edge, upstream_state in upstream_states.items():

        # ensure we are working with populated result objects
        if edge.key in state.cached_inputs:
            upstream_state._result = state.cached_inputs[edge.key]

        # if the upstream was mapped and the edge is flattened, we need
        # to process the mapped children (which could be futures) into a
        # flat structure
        if upstream_state.is_mapped() and edge.flattened:
            mapped_children[edge.upstream_task] = flatten_mapped_children(
                mapped_children=mapped_children[edge.upstream_task], executor=executor
            )

    # infinite loop, if upstream_states has any entries
    while True and upstream_states:
        i = next(counter)
        states = {}
        try:

            for edge, upstream_state in upstream_states.items():

                # if the edge is not mapped over, then we take its state
                if not edge.mapped:
                    states[edge] = upstream_state

                # if the edge is mapped and the upstream state is Mapped, then we are mapping
                # over a mapped task. In this case, we take the appropriately-indexed upstream
                # state from the upstream tasks's `Mapped.map_states` array.
                # Note that these "states" might actually be futures at this time; we aren't
                # blocking until they finish.
                elif edge.mapped and upstream_state.is_mapped():
                    states[edge] = mapped_children[edge.upstream_task][i]  # type: ignore

                # Otherwise, we are mapping over the result of a "vanilla" task. In this
                # case, we create a copy of the upstream state but set the result to the
                # appropriately-indexed item from the upstream task's `State.result`
                # array.
                else:
                    states[edge] = copy.copy(upstream_state)

                    # if the current state is already Mapped, then we might be executing
                    # a re-run of the mapping pipeline. In that case, the upstream states
                    # might not have `result` attributes.
                    # Therefore, we only try to get a result if EITHER this task's
                    # state is not already mapped OR the upstream result is not None.
                    if (
                        not state.is_mapped()
                        or upstream_state._result != prefect.engine.result.NoResult
                    ):
                        if not hasattr(upstream_state.result, "__getitem__"):
                            value = None
                        else:
                            value = upstream_state.result[i]
                        upstream_result = upstream_state._result.from_value(value)  # type: ignore
                        states[edge].result = upstream_result
                        if state.map_states and i >= len(state.map_states):  # type: ignore
                            raise IndexError()
                    elif state.is_mapped():
                        if i >= len(state.map_states):  # type: ignore
                            raise IndexError()

            # only add this iteration if we made it through all iterables
            map_upstream_states.append(states)

        # index error means we reached the end of the shortest iterable
        except IndexError:
            break

    return map_upstream_states


def _build_flattened_state(state: "State", index: int) -> "State":
    """Helper function for `flatten_upstream_state`"""
    new_state = copy.copy(state)
    new_state.result = state._result.from_value(state.result[index])  # type: ignore
    return new_state


def flatten_upstream_state(upstream_state: "State") -> "State":
    """
    Given an upstream state, returns its result as a flattened list. If
    flattening fails, the object is returned unmodified.
    """
    try:
        # attempt to unnest
        flattened_result = [y for x in upstream_state.result for y in x]
    except TypeError:
        return upstream_state

    new_state = copy.copy(upstream_state)
    new_state.result = new_state._result.from_value(flattened_result)  # type: ignore
    return new_state


def flatten_mapped_children(
    mapped_children: List["State"],
    executor: "prefect.executors.Executor",
) -> List["State"]:
    counts = executor.wait(
        [executor.submit(lambda c: len(c._result.value), c) for c in mapped_children]
    )
    new_states = []

    for child, count in zip(mapped_children, counts):
        new_states.append(
            [executor.submit(_build_flattened_state, child, i) for i in range(count)]
        )

    flattened_states = [i for s in new_states for i in s]
    return flattened_states



"""
This module contains a collection of tasks for interacting with Databricks resources.
"""

from prefect.tasks.databricks.databricks_submitjob import DatabricksSubmitRun
from prefect.tasks.databricks.databricks_submitjob import DatabricksRunNow



import json
import time

import click
from tabulate import tabulate

from prefect.client import Client
from prefect.utilities.graphql import EnumValue, with_args


@click.group(hidden=True)
def run():
    """
    Run Prefect flows.

    \b
    Usage:
        $ prefect run [STORAGE/PLATFORM]

    \b
    Arguments:
        flow    Run a flow with a backend API

    \b
    Examples:
        $ prefect run flow --name Test-Flow --project My-Project -ps '{"my_param": 42}'
        Flow Run: https://cloud.prefect.io/myslug/flow-run/2ba3rrfd-411c-4d99-bb2a-f64a6dea78f9

    \b
        $ prefect run flow --name Test-Flow --project My-Project --watch
        Flow Run: https://localhost:8080/flow-run/2ba3rrfd-411c-4d99-bb2a-f64a6dea78f9
        Scheduled -> Submitted -> Running -> Success
    """


@run.command(hidden=True)
@click.option(
    "--id", "-i", required=False, help="The id of a flow to run.", hidden=True
)
@click.option(
    "--version-group-id",
    required=False,
    help="The id of a flow version group to run.",
    hidden=True,
)
@click.option(
    "--name", "-n", required=False, help="The name of a flow to run.", hidden=True
)
@click.option(
    "--project",
    "-p",
    required=False,
    help="The project that contains the flow.",
    hidden=True,
)
@click.option("--version", "-v", type=int, help="A flow version to run.", hidden=True)
@click.option(
    "--parameters-file",
    "-pf",
    help="A parameters JSON file.",
    hidden=True,
    type=click.Path(exists=True),
)
@click.option(
    "--parameters-string", "-ps", help="A parameters JSON string.", hidden=True
)
@click.option("--run-name", "-rn", help="A name to assign for this run.", hidden=True)
@click.option("--context", "-c", help="A context JSON string.", hidden=True)
@click.option(
    "--watch",
    "-w",
    is_flag=True,
    help="Watch current state of the flow run.",
    hidden=True,
)
@click.option(
    "--label",
    "labels",
    help="A list of labels to apply to the flow run",
    hidden=True,
    multiple=True,
)
@click.option(
    "--logs", "-l", is_flag=True, help="Live logs of the flow run.", hidden=True
)
@click.option(
    "--no-url",
    is_flag=True,
    help="Only output flow run id instead of link.",
    hidden=True,
)
def flow(
    id,
    version_group_id,
    name,
    project,
    version,
    parameters_file,
    parameters_string,
    run_name,
    context,
    watch,
    labels,
    logs,
    no_url,
):
    """
    Run a flow that is registered to the Prefect API

    \b
    Options:
        --id, -i                    TEXT        The ID of a flow to run
        --version-group-id          TEXT        The ID of a flow version group to run
        --name, -n                  TEXT        The name of a flow to run
        --project, -p               TEXT        The name of a project that contains the flow
        --version, -v               INTEGER     A flow version to run
        --parameters-file, -pf      FILE PATH   A filepath of a JSON file containing
                                                parameters
        --parameters-string, -ps    TEXT        A string of JSON parameters (note: to ensure these are
                                                parsed correctly, it is best to include the full payload
                                                within single quotes)
        --run-name, -rn             TEXT        A name to assign for this run
        --context, -c               TEXT        A string of JSON key / value pairs to include in context
                                                (note: to ensure these are parsed correctly, it is best
                                                to include the full payload within single quotes)
        --watch, -w                             Watch current state of the flow run, stream
                                                output to stdout
        --label                     TEXT        Set labels on the flow run; use multiple times to set
                                                multiple labels.
        --logs, -l                              Get logs of the flow run, stream output to
                                                stdout
        --no-url                                Only output the flow run id instead of a
                                                link

    \b
    Either `id`, `version-group-id`, or both `name` and `project` must be provided to run a flow.

    \b
    If both `--parameters-file` and `--parameters-string` are provided then the values
    passed in through the string will override the values provided from the file.

    \b
    e.g.
    File contains:  {"a": 1, "b": 2}
    String:         '{"a": 3}'
    Parameters passed to the flow run: {"a": 3, "b": 2}

    \b
    Example:
        $ prefect run flow -n "Test-Flow" -p "My Project" -ps '{"my_param": 42}'
        Flow Run: https://cloud.prefect.io/myslug/flow-run/2ba3rrfd-411c-4d99-bb2a-f64a6dea78f9
    """
    if not id and not (name and project) and not version_group_id:
        click.secho(
            "A flow ID, version group ID, or a combination of flow name and project must be provided.",
            fg="red",
        )
        return

    if sum(map(bool, (id, version_group_id, name))) != 1:
        click.secho(
            "Only one of flow ID, version group ID, or a name/project combination can be provided.",
            fg="red",
        )
        return

    if watch and logs:
        click.secho(
            "Streaming state and logs not currently supported together.", fg="red"
        )
        return

    if labels == ():
        labels = None

    client = Client()
    flow_id = id
    if not flow_id and not version_group_id:
        where_clause = {
            "_and": {
                "name": {"_eq": name},
                "version": {"_eq": version},
                "project": {"name": {"_eq": project}},
            }
        }

        query = {
            "query": {
                with_args(
                    "flow",
                    {
                        "where": where_clause,
                        "order_by": {
                            "name": EnumValue("asc"),
                            "version": EnumValue("desc"),
                        },
                        "distinct_on": EnumValue("name"),
                    },
                ): {"id": True}
            }
        }

        result = client.graphql(query)

        flow_data = result.data.flow

        if flow_data:
            flow_id = flow_data[0].id
        else:
            click.secho("{} not found".format(name), fg="red")
            return

    # Load parameters from file if provided
    file_params = {}
    if parameters_file:
        with open(parameters_file) as params_file:
            file_params = json.load(params_file)

    # Load parameters from string if provided
    string_params = {}
    if parameters_string:
        string_params = json.loads(parameters_string)

    if context:
        context = json.loads(context)
    flow_run_id = client.create_flow_run(
        flow_id=flow_id,
        version_group_id=version_group_id,
        context=context,
        labels=labels,
        parameters={**file_params, **string_params},
        run_name=run_name,
    )

    if no_url:
        click.echo("Flow Run ID: {}".format(flow_run_id))
    else:
        flow_run_url = client.get_cloud_url("flow-run", flow_run_id)
        click.echo("Flow Run: {}".format(flow_run_url))

    if watch:
        current_states = []
        while True:
            query = {
                "query": {
                    with_args("flow_run_by_pk", {"id": flow_run_id}): {
                        with_args(
                            "states",
                            {"order_by": {EnumValue("timestamp"): EnumValue("asc")}},
                        ): {"state": True, "timestamp": True}
                    }
                }
            }

            result = client.graphql(query)

            # Filter through retrieved states and output in order
            for state_index in result.data.flow_run_by_pk.states:
                state = state_index.state
                if state not in current_states:
                    if state != "Success" and state != "Failed":
                        click.echo("{} -> ".format(state), nl=False)
                    else:
                        click.echo(state)
                        return flow_run_id

                    current_states.append(state)

            time.sleep(3)

    if logs:
        all_logs = []

        log_query = {
            with_args(
                "logs", {"order_by": {EnumValue("timestamp"): EnumValue("asc")}}
            ): {"timestamp": True, "message": True, "level": True},
            "start_time": True,
            "state": True,
        }

        query = {
            "query": {
                with_args(
                    "flow_run",
                    {
                        "where": {"id": {"_eq": flow_run_id}},
                        "order_by": {EnumValue("start_time"): EnumValue("desc")},
                    },
                ): log_query
            }
        }

        while True:
            result = client.graphql(query)

            flow_run = result.data.flow_run
            if not flow_run:
                click.secho("{} not found".format(flow_run_id), fg="red")
                return

            new_run = flow_run[0]
            logs = new_run.logs
            output = []

            for i in logs:
                if [i.timestamp, i.level, i.message] not in all_logs:

                    if not len(all_logs):
                        click.echo(
                            tabulate(
                                [[i.timestamp, i.level, i.message]],
                                headers=["TIMESTAMP", "LEVEL", "MESSAGE"],
                                tablefmt="plain",
                                numalign="left",
                                stralign="left",
                            )
                        )
                        all_logs.append([i.timestamp, i.level, i.message])
                        continue

                    output.append([i.timestamp, i.level, i.message])
                    all_logs.append([i.timestamp, i.level, i.message])

            if output:
                click.echo(
                    tabulate(output, tablefmt="plain", numalign="left", stralign="left")
                )

            if new_run.state == "Success" or new_run.state == "Failed":
                return flow_run_id

            time.sleep(3)

    return flow_run_id



"""
The tasks in this module can be used to handle file compression like zipping or
unzipping files.
"""
import os
import sys
import zipfile
from pathlib import Path
from typing import Any, List, Union
from zipfile import ZipFile

from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs


class Unzip(Task):
    """
    Task for unzipping data.

    Args:
        - zip_path (Union[str, Path], optional): the path to the zip file
        - extract_dir (Union[str, Path], optional): directory to extract the zip file into.
            If not provided, the current working directory will be used. This directory must
            already exist.
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        zip_path: Union[str, Path] = "",
        extract_dir: Union[str, Path] = "",
        **kwargs: Any,
    ):
        self.zip_path = zip_path
        self.extract_dir = extract_dir
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "zip_path",
        "extract_dir",
    )
    def run(
        self,
        zip_path: Union[str, Path] = "",
        extract_dir: Union[str, Path] = "",
        password: Union[bytes, str] = None,
    ) -> Path:
        """
        Task run method.

        Args:
            - zip_path (Union[str, Path], optional): the path to the zip file
            - extract_dir (Union[str, Path], optional): directory to extract the zip file into.
                If not provided, the current working directory will be used. This directory must
                already exist.
            - password (Union[bytes, str], optional): password for unzipping a password-protected
                zip file. If a `str`, will be utf-8 encoded.

        Returns:
            - Path: path to the extracted directory.
        """
        if not zip_path:
            raise ValueError("No `zip_path` provided")

        zip_path = Path(zip_path)
        extract_dir = Path(extract_dir or Path.cwd())

        if not zip_path.exists():
            raise ValueError(f"`zip_path` '{zip_path}' not found")

        if not zipfile.is_zipfile(zip_path):
            raise TypeError(f"`zip_path` '{zip_path}' is not a zip file")

        self.logger.info(f"Extracting '{zip_path}' to '{extract_dir}'")

        if password and isinstance(password, str):
            password = password.encode("utf-8")
        with ZipFile(zip_path, "r") as zip:
            zip.extractall(extract_dir, pwd=password)

        return extract_dir


class Zip(Task):
    """
    Task to create a zip archive.

    Args:
        - source_path (Union[str, Path, List[str], List[Path]], optional): path or paths to compress
             into a single zip archive.
        - zip_path (Union[str, Path], optional): path to the output archive file. Any parent
            directories of `zip_path` must already exist.
        - compression_method (str, optional): the compression method to use. Options are
            "deflate", "store", "bzip2", and "lzma". Defaults to `"deflate"`.
        - compression_level (int, optional): Compression level to use,
            see https://docs.python.org/3/library/zipfile.html#zipfile.ZipFile for more info.
            Python 3.7+ only.
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor.
    """

    def __init__(
        self,
        source_path: Union[str, Path, List[str], List[Path]] = "",
        zip_path: Union[str, Path] = "",
        compression_method: str = "deflate",
        compression_level: int = None,
        **kwargs: Any,
    ):
        self.source_path = source_path
        self.zip_path = zip_path

        if compression_method == "store":
            self.compression = zipfile.ZIP_STORED
        elif compression_method == "deflate":
            self.compression = zipfile.ZIP_DEFLATED
        elif compression_method == "bzip2":
            self.compression = zipfile.ZIP_BZIP2
        elif compression_method == "lzma":
            self.compression = zipfile.ZIP_LZMA
        else:
            raise ValueError(
                f"Compression method {compression_method} is not supported!"
            )

        self.compression_level = compression_level
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "source_path",
        "zip_path",
    )
    def run(
        self,
        source_path: Union[str, Path, List[str], List[Path]] = "",
        zip_path: Union[str, Path] = "",
    ) -> None:
        """
        Task run method.

        Args:
            - source_path (Union[str, Path, List[str], List[Path]], optional): path or paths to compress
                into a single zip archive.
            - zip_path (Union[str, Path], optional): path to the output archive file. Any parent
                directories of `zip_path` must already exist.
        """

        if not source_path:
            raise ValueError("No `source_path` provided.")
        if not zip_path:
            raise ValueError("No `zip_path` provided.")

        zip_path = Path(zip_path)

        if not isinstance(source_path, list):
            source_path = [Path(source_path)]
        else:
            source_path = [Path(sp) for sp in source_path]

        if sys.version_info >= (3, 7):
            # compresslevel was added in 3.7
            zip_file = ZipFile(
                zip_path,
                "w",
                compression=self.compression,
                compresslevel=self.compression_level,
            )
        else:
            zip_file = ZipFile(
                zip_path,
                "w",
                compression=self.compression,
            )

        with zip_file:
            self.logger.info(f"Creating zip archive {zip_path}")
            for f in source_path:
                self.logger.info(f"Adding {f.name} to archive")
                if f.is_file():
                    zip_file.write(f, arcname=f.name)
                else:
                    for root, dirs, files in os.walk(f):
                        root_path = Path(root)
                        for file in files:
                            fp = root_path.joinpath(file)
                            zip_file.write(fp, arcname=fp.relative_to(f.parent))
